from __future__ import absolute_import

import re
from datetime import datetime, timedelta
from unittest.mock import ANY, Mock, patch

import pytest
import pytz
from common.mail import campaign_ids
from common.models import DailyActivity, Student, Teacher
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly
from common.tests.utils.student import (
    create_independent_student_directly,
    create_school_student_directly,
    generate_independent_student_details,
)
from common.tests.utils.teacher import generate_details, signup_teacher_directly
from django.core import mail
from django.test import Client, TestCase
from django.urls import reverse, reverse_lazy

from portal.helpers.ratelimit import get_ratelimit_count_for_user
from portal.views.login import has_user_lockout_expired


class TestRatelimit(TestCase):
    def setUp(self) -> None:
        self.client = Client()

    def _teacher_login(self, username, password):
        return self.client.post(
            reverse("teacher_login"),
            {"auth-username": username, "auth-password": password, "teacher_login_view-current_step": "auth"},
        )

    def _student_login(self, username, password):
        return self.client.post(reverse("independent_student_login"), {"username": username, "password": password})

    def _student_school_login(self, access_code, student_name, student_password):
        return self.client.post(
            reverse("student_login", kwargs={"access_code": access_code}),
            {"username": student_name, "password": student_password},
            follow=True,
        )

    def _teacher_update_account_bad_request(self) -> None:
        """
        Performs a failed POST request to the teacher update account form, as it uses
        a bad password.
        """
        self.client.post(
            reverse("dashboard"),
            {
                "first_name": "Test",
                "last_name": "User",
                "email": "",
                "password": "",
                "confirm_password": "",
                "current_password": "bad_password",
                "update_account": "",
            },
        )

    def _independent_student_edit_account_bad_request(self) -> None:
        """
        Performs a failed POST request to the independent student edit account form, as
        it uses a bad password.
        """
        self.client.post(
            reverse("independent_edit_account"),
            {
                "name": "Test User",
                "email": "",
                "password": "",
                "confirm_password": "",
                "current_password": "bad_password",
            },
        )

    def _reset_password_request(self, email):
        return self.client.post(
            reverse("teacher_password_reset"), {"email": email, "g-recaptcha-response": "something"}
        )

    def _reset_password(self, url, new_password):
        return self.client.post(url, {"new_password1": new_password, "new_password2": new_password})

    def _is_user_blocked(self, model: Teacher or Student, username: str, access_code: str = None) -> bool:
        """
        Checks if a Teacher or a Student object is blocked, by checking if they
        have a blocked_time value, and if so, if it the lockout has expired or not.
        :param model: The model Class to be checked against.
        :param username: The username of the Teacher or Student.
        :return: Whether or not the model object is marked as blocked.
        """
        user = (
            model.objects.get(new_user__username=username)
            if not access_code
            else model.objects.get(new_user__first_name=username, class_field__access_code=access_code)
        )
        if user.blocked_time:
            return not has_user_lockout_expired(user)
        else:
            return False

    def _block_user(self, model: Teacher or Student, username: str, access_code=None) -> None:
        """
        Finds the Teacher or Student corresponding to the username, and sets it as
        blocked and sets the blocked date to now.
        :param model: The model Class to be checked against.
        :param username: The username of the Teacher or Student.
        """

        user = (
            model.objects.get(new_user__username=username)
            if access_code is None
            else model.objects.get(new_user__first_name=username, class_field__access_code=access_code)
        )

        user.blocked_time = datetime.now(tz=pytz.utc)
        user.save()

    def test_teacher_login_ratelimit(self):
        """
        Given a teacher,
        When they perform 6 failed login attempts,
        Then on the 6th one, the teacher should be blocked.
        """
        email, password = signup_teacher_directly()

        for i in range(5):
            _ = self._teacher_login(email, "bad_password")

            assert not self._is_user_blocked(Teacher, email)

        _ = self._teacher_login(email, "bad_password")

        assert self._is_user_blocked(Teacher, email)

    def test_student_login_ratelimit(self):
        """
        Given a student,
        When they perform 6 failed login attempts,
        Then on the 6th one, the student should be blocked.
        """
        teacher_email, teacher_password = signup_teacher_directly()
        school = create_organisation_directly(teacher_email)
        klass, klass_name, klass_access_code = create_class_directly(teacher_email)
        student_name, student_password, student = create_school_student_directly(klass_access_code)

        for i in range(10):
            response = self._student_school_login(klass_access_code, student_name, "bad_password")

            assert not self._is_user_blocked(Student, student_name, klass_access_code)

        _ = self._student_school_login(klass_access_code, student_name, "bad_password")

        assert self._is_user_blocked(Student, student_name, klass_access_code)
        student = Student.objects.get(id=student.id)
        current_student = Student.objects.get(
            new_user__first_name=student_name, class_field__access_code=klass_access_code
        )

        # now check if teacher can unlock it, both ways :)
        url = reverse_lazy("teacher_class_password_reset", kwargs={"access_code": klass_access_code})
        data = {"transfer_students": [[current_student.id]]}
        c = Client()

        c.login(username=teacher_email, password=teacher_password)
        c.post(url, data)
        assert not self._is_user_blocked(Student, student_name, klass_access_code)

        # now block again and test the edit by student method
        self._block_user(Student, student_name, klass_access_code)
        assert self._is_user_blocked(Student, student_name, klass_access_code)
        url = reverse_lazy("teacher_edit_student", kwargs={"pk": current_student.id})
        strong_password = "£EDCVFR$5tgbnhy6"
        data = {"password": strong_password, "confirm_password": strong_password, "set_password": ""}

        c.post(url, data)
        assert not self._is_user_blocked(Student, student_name, klass_access_code)
        c.logout()
        student = Student.objects.get(id=student.id)
        self._student_school_login(klass_access_code, student_name, "password1")
        student = Student.objects.get(id=student.id)
        assert not self._is_user_blocked(Student, student_name, klass_access_code)

    def test_independent_student_login_ratelimit(self):
        """
        Given an independent student,
        When they perform 6 failed login attempts,
        Then on the 6th one, the independent student should be blocked.
        """
        username, password, _ = create_independent_student_directly()

        for i in range(5):
            _ = self._student_login(username, "bad_password")

            assert not self._is_user_blocked(Student, username)

        _ = self._student_login(username, "bad_password")

        assert self._is_user_blocked(Student, username)

    def test_teacher_update_account_ratelimit(self):
        """
        Given a teacher with a school, class and student,
        When they perform 6 failed attempts to update their account details,
        Then on the 6th one, the teacher should be blocked.
        """
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        klass, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        _ = self._teacher_login(email, password)

        for i in range(5):
            self._teacher_update_account_bad_request()

            assert not self._is_user_blocked(Teacher, email)

        self._teacher_update_account_bad_request()

        assert self._is_user_blocked(Teacher, email)

    def test_independent_student_update_account_ratelimit(self):
        """
        Given an independent student,
        When they perform 6 failed attempts to edit their account details,
        Then on the 6th one, the independent student should be blocked.
        """
        username, password, _ = create_independent_student_directly()

        _ = self._student_login(username, password)

        for i in range(5):
            self._independent_student_edit_account_bad_request()

            assert not self._is_user_blocked(Student, username)

        self._independent_student_edit_account_bad_request()

        assert self._is_user_blocked(Student, username)

    def test_teacher_account_lockout(self):
        """
        Given a blocked teacher,
        When they try to login,
        They should not be redirected anywhere.
        Then, given they've waited the time to unlock their account,
        When they try to login,
        They should be able to login, and not be blocked anymore.
        """
        email, password = signup_teacher_directly()

        self._block_user(Teacher, email)

        login_response = self._teacher_login(email, password)

        # Check for a 200, instead of the usual 302.
        assert login_response.status_code == 200

        # Manually change the blocked date to over 24 hours ago to emulate waiting.
        teacher = Teacher.objects.get(new_user__username=email)
        teacher.blocked_time = datetime.now(tz=pytz.utc) - timedelta(hours=24)
        teacher.save()

        login_response = self._teacher_login(email, password)

        assert login_response.status_code == 302
        assert not self._is_user_blocked(Teacher, email)

    def test_student_account_lockout(self):
        """
        Given a blocked student,
        When they try to login,
        They should not be redirected anywhere.
        Then, given they've waited the time to unlock their account,
        When they try to login,
        They should be able to login, and not be blocked anymore.
        """
        username, password, _ = create_independent_student_directly()

        self._block_user(Student, username)

        login_response = self._student_login(username, password)

        # Check for a 200, instead of the usual 302.
        assert login_response.status_code == 200

        # Manually change the blocked date to over 24 hours ago to emulate waiting.
        student = Student.objects.get(new_user__username=username)
        student.blocked_time = datetime.now(tz=pytz.utc) - timedelta(hours=24)
        student.save()

        login_response = self._student_login(username, password)

        assert login_response.status_code == 302
        assert not self._is_user_blocked(Student, username)

    def test_successful_request_resets_cache(self):
        """
        Given a teacher who's performed some failed login attempts,
        When they login successfully,
        Then the ratelimit cache should be reset.
        """
        email, password = signup_teacher_directly()

        # Fail login 3 times
        for i in range(3):
            self._teacher_login(email, "bad_password")

        assert get_ratelimit_count_for_user(email) == 3

        # Login successfully
        self._teacher_login(email, password)

        assert get_ratelimit_count_for_user(email) is None

        self.client.logout()

        # Fail login again one more time
        self._teacher_login(email, "bad_password")

        assert get_ratelimit_count_for_user(email) == 1

    @patch("portal.forms.registration.send_dotdigital_email")
    def test_teacher_reset_password_unblocks_user(self, mock_send_dotdigital_email: Mock):
        """
        Given a blocked teacher,
        When they reset they password,
        They should be unblocked and be able to login.
        """
        email, password = signup_teacher_directly()

        self._block_user(Teacher, email)

        login_response = self._teacher_login(email, password)

        # Check for a 200, instead of the usual 302.
        assert login_response.status_code == 200

        # Ask for reset password link
        self._reset_password_request(email)

        mock_send_dotdigital_email.assert_called_once_with(
            campaign_ids["reset_password"], ANY, personalization_values=ANY
        )

        reset_password_url = mock_send_dotdigital_email.call_args.kwargs["personalization_values"][
            "RESET_PASSWORD_LINK"
        ]

        new_password = "AnotherPassword12!"

        self._reset_password(reset_password_url, new_password)

        login_response = self._teacher_login(email, new_password)

        assert login_response.status_code == 302
        assert not self._is_user_blocked(Teacher, email)

    def test_lockout_reset_tracking(self):
        old_date = datetime.now() - timedelta(days=1)
        old_daily_activity = DailyActivity(date=old_date)
        old_daily_activity.save()
        teacher_email, teacher_password = signup_teacher_directly()
        indy_email, indy_password, student = create_independent_student_directly()
        create_organisation_directly(teacher_email)

        self._block_user(Teacher, teacher_email)
        self._block_user(Student, indy_email)

        # check teacher response for resetting password
        url = reverse_lazy("teacher_password_reset")
        data = {"email": teacher_email}

        c = Client()

        response = c.post(url, data=data)
        old_daily_activity = DailyActivity.objects.get(date=old_date)
        current_daily_activity = DailyActivity.objects.get(date=datetime.now())

        assert response.status_code == 200
        assert old_daily_activity.teacher_lockout_resets == 0
        assert current_daily_activity.teacher_lockout_resets == 1
        # now check the indy student

        url = reverse_lazy("student_password_reset")
        data = {"email": indy_email}
        c = Client()

        response = c.post(url, data=data)
        old_daily_activity = DailyActivity.objects.get(date=old_date)
        current_daily_activity = DailyActivity.objects.get(date=datetime.now())

        assert response.status_code == 200
        assert old_daily_activity.indy_lockout_resets == 0
        assert current_daily_activity.indy_lockout_resets == 1
        # finally check the school student

        # method 1
        _, _, klass_access_code = create_class_directly(teacher_email)
        student_name, _, student = create_school_student_directly(klass_access_code)

        self._block_user(Student, student_name, access_code=klass_access_code)

        c = Client()
        c.login(username=teacher_email, password=teacher_password)

        url = reverse_lazy("teacher_edit_student", kwargs={"pk": student.id})
        strong_password = "£EDCVFR$5tgb"
        data = {"password": strong_password, "confirm_password": strong_password, "set_password": ""}

        response = c.post(url, data)
        old_daily_activity = DailyActivity.objects.get(date=old_date)
        current_daily_activity = DailyActivity.objects.get(date=datetime.now())

        assert response.status_code == 200
        assert old_daily_activity.school_student_lockout_resets == 0
        assert current_daily_activity.school_student_lockout_resets == 1

        # method 2
        self._block_user(Student, student_name, access_code=klass_access_code)
        url = reverse_lazy("teacher_class_password_reset", kwargs={"access_code": klass_access_code})
        data = {"transfer_students": [[student.id]]}

        response = c.post(url, data)
        old_daily_activity = DailyActivity.objects.get(date=old_date)
        current_daily_activity = DailyActivity.objects.get(date=datetime.now())

        assert response.status_code == 200
        assert old_daily_activity.school_student_lockout_resets == 0
        assert current_daily_activity.school_student_lockout_resets == 2


@patch("common.helpers.emails.send_dotdigital_email")
@pytest.mark.django_db
def test_teacher_already_registered_email(mock_send_dotdigital_email: Mock, client):
    first_name, last_name, email, password = generate_details()
    register_url = reverse("register")
    data = {
        "teacher_signup-teacher_first_name": first_name,
        "teacher_signup-teacher_last_name": last_name,
        "teacher_signup-teacher_email": email,
        "teacher_signup-consent_ticked": "on",
        "teacher_signup-teacher_password": password,
        "teacher_signup-teacher_confirm_password": password,
        "g-recaptcha-response": "something",
    }

    # Register the teacher first time, there should be a registration email
    client.post(register_url, data)
    mock_send_dotdigital_email.assert_called_once_with(campaign_ids["verify_new_user"], ANY, personalization_values=ANY)

    # Register with the same email again, there should also be an already registered email
    client.post(register_url, data)
    assert len(mail.outbox) == 1

    # Register with the same email one more time, there shouldn't be any new emails
    client.post(register_url, data)
    assert len(mail.outbox) == 1


@patch("common.helpers.emails.send_dotdigital_email")
@pytest.mark.django_db
def test_independent_student_already_registered_email(mock_send_dotdigital_email: Mock, client):
    name, username, email_address, password = generate_independent_student_details()
    register_url = reverse("register")
    data = {
        "independent_student_signup-date_of_birth_day": 7,
        "independent_student_signup-date_of_birth_month": 10,
        "independent_student_signup-date_of_birth_year": 1997,
        "independent_student_signup-name": name,
        "independent_student_signup-email": email_address,
        "independent_student_signup-consent_ticked": "on",
        "independent_student_signup-password": password,
        "independent_student_signup-confirm_password": password,
        "g-recaptcha-response": "something",
    }

    # Register the independent student first time, there should be a registration email
    client.post(register_url, data)
    mock_send_dotdigital_email.assert_called_once_with(campaign_ids["verify_new_user"], ANY, personalization_values=ANY)

    # Register with the same email again, there should also be an already registered email
    client.post(register_url, data)
    assert len(mail.outbox) == 1

    # Reset mock and register with the same email one more time, there shouldn't be any new emails
    client.post(register_url, data)
    assert len(mail.outbox) == 1



# Generated by Django 2.0 on 2020-10-28 17:36

import datetime
from uuid import uuid4

from django.db import migrations
from django.utils import timezone


def verify_portaladmin(apps, schema_editor):
    """
    This migration is so that we can still log in using the portaladmin User, as the
    login form now requires email verification.
    """
    User = apps.get_model("auth", "User")
    EmailVerification = apps.get_model("common", "EmailVerification")

    portaladmin = User.objects.get(username="portaladmin")

    EmailVerification.objects.create(
        user=portaladmin,
        email=portaladmin.email,
        token=uuid4().hex[:30],
        expiry=timezone.now() + datetime.timedelta(hours=1),
        verified=True,
    )


def revert_portaladmin_verification(apps, schema_editor):
    User = apps.get_model("auth", "User")
    EmailVerification = apps.get_model("common", "EmailVerification")

    portaladmin = User.objects.get(username="portaladmin")
    portaladmin_verification = EmailVerification.objects.get(user=portaladmin)
    portaladmin_verification.delete()


class Migration(migrations.Migration):

    dependencies = [("portal", "0061_make_portaladmin_teacher")]

    operations = [
        migrations.RunPython(verify_portaladmin, revert_portaladmin_verification)
    ]



from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [("portal", "0051_add_missing_ev_records")]

    operations = [
        migrations.RemoveField(model_name="emailverification", name="new_user")
    ]



import csv
import json
from datetime import datetime, timedelta
from enum import Enum
from functools import partial, wraps
from uuid import uuid4

import pytz
from common.helpers.emails import send_verification_email
from common.helpers.generators import (
    generate_access_code,
    generate_login_id,
    generate_password,
    get_hashed_login_id,
)
from common.models import (
    Class,
    DailyActivity,
    JoinReleaseStudent,
    Student,
    Teacher,
    TotalActivity,
)
from common.permissions import check_teacher_authorised, logged_in_as_teacher
from django.contrib import messages as messages
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.models import User
from django.contrib.staticfiles.storage import staticfiles_storage
from django.db.models import F
from django.forms.formsets import formset_factory
from django.http import Http404, HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.views.decorators.http import require_POST
from game.models import Level
from game.views.level_selection import get_blockly_episodes, get_python_episodes
from reportlab.lib.colors import black, red
from reportlab.lib.pagesizes import A4
from reportlab.lib.utils import ImageReader
from reportlab.pdfgen import canvas

from portal.forms.teach import (
    BaseTeacherDismissStudentsFormSet,
    BaseTeacherMoveStudentsDisambiguationFormSet,
    ClassCreationForm,
    ClassEditForm,
    ClassLevelControlForm,
    ClassMoveForm,
    StudentCreationForm,
    TeacherDismissStudentsForm,
    TeacherEditStudentForm,
    TeacherMoveStudentDisambiguationForm,
    TeacherMoveStudentsDestinationForm,
    TeacherSetStudentPass,
)
from portal.helpers.ratelimit import clear_ratelimit_cache_for_user
from portal.views.registration import handle_reset_password_tracking

STUDENT_PASSWORD_LENGTH = 6
REMINDER_CARDS_PDF_ROWS = 8
REMINDER_CARDS_PDF_COLUMNS = 1
REMINDER_CARDS_PDF_WARNING_TEXT = "Please ensure students keep login details in a secure place"


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_onboarding_create_class(request):
    """
    Onboarding view for creating a class (and organisation if there isn't one, yet)
    """
    teacher = request.user.new_teacher
    requests = Student.objects.filter(pending_class_request__teacher=teacher, new_user__is_active=True)

    if not teacher.school:
        return HttpResponseRedirect(reverse_lazy("onboarding-organisation"))

    if request.method == "POST":
        form = ClassCreationForm(request.POST, teacher=teacher)
        if form.is_valid():
            created_class = create_class(form, teacher)
            messages.success(
                request,
                "The class '{className}' has been created successfully.".format(className=created_class.name),
            )
            return HttpResponseRedirect(
                reverse_lazy(
                    "onboarding-class",
                    kwargs={"access_code": created_class.access_code},
                )
            )
    else:
        form = ClassCreationForm(teacher=teacher)

    classes = Class.objects.filter(teacher=teacher)

    return render(
        request,
        "portal/teach/onboarding_classes.html",
        {"form": form, "classes": classes, "requests": requests},
    )


def create_class(form, class_teacher, class_creator=None):
    classmate_progress = bool(form.cleaned_data["classmate_progress"])
    klass = Class.objects.create(
        name=form.cleaned_data["class_name"],
        teacher=class_teacher,
        access_code=generate_access_code(),
        classmates_data_viewable=classmate_progress,
        created_by=class_teacher if class_creator is None else class_creator,
    )
    return klass


def generate_student_url(request, student, login_id):
    return request.build_absolute_uri(
        reverse(
            "student_direct_login",
            kwargs={"user_id": student.new_user.id, "login_id": login_id},
        )
    )


def process_edit_class(request, access_code, onboarding_done, next_url):
    """
    Handles student creation both during onboarding or on the class page
    """
    klass = get_object_or_404(Class, access_code=access_code)
    teacher = request.user.new_teacher
    students = Student.objects.filter(class_field=klass, new_user__is_active=True).order_by("new_user__first_name")

    check_teacher_authorised(request, klass.teacher)

    if request.method == "POST":
        new_students_form = StudentCreationForm(klass, request.POST)
        if new_students_form.is_valid():
            students_info = []
            for name in new_students_form.strippedNames:
                password = generate_password(STUDENT_PASSWORD_LENGTH)

                # generate uuid for url and store the hashed
                login_id, hashed_login_id = generate_login_id()

                new_student = Student.objects.schoolFactory(
                    klass=klass,
                    name=name,
                    password=password,
                    login_id=hashed_login_id,
                )

                TotalActivity.objects.update(student_registrations=F("student_registrations") + 1)

                login_url = generate_student_url(request, new_student, login_id)
                students_info.append(
                    {
                        "id": new_student.new_user.id,
                        "name": name,
                        "password": password,
                        "login_url": login_url,
                    }
                )

            return render(
                request,
                "portal/teach/onboarding_print.html",
                {
                    "class": klass,
                    "students_info": students_info,
                    "onboarding_done": onboarding_done,
                    "query_data": json.dumps(students_info),
                    "class_url": request.build_absolute_uri(
                        reverse(
                            "student_login",
                            kwargs={"access_code": klass.access_code},
                        )
                    ),
                },
            )
    else:
        new_students_form = StudentCreationForm(klass)

    classes = Class.objects.filter(teacher=teacher)

    return render(
        request,
        next_url,
        {
            "class": klass,
            "classes": classes,
            "students": students,
            "new_students_form": new_students_form,
            "num_students": len(students),
        },
    )


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_onboarding_edit_class(request, access_code):
    """
    Adding students to a class during the onboarding process
    """
    return process_edit_class(
        request,
        access_code,
        onboarding_done=False,
        next_url="portal/teach/onboarding_students.html",
    )


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_view_class(request, access_code):
    """
    Adding students to a class after the onboarding process has been completed
    """
    return process_edit_class(
        request,
        access_code,
        onboarding_done=True,
        next_url="portal/teach/class.html",
    )


@require_POST
@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_delete_class(request, access_code):
    klass = get_object_or_404(Class, access_code=access_code)

    # check user authorised to see class
    check_teacher_authorised(request, klass.teacher)

    if Student.objects.filter(class_field=klass, new_user__is_active=True).exists():
        messages.info(
            request,
            "This class still has students, please remove or delete them all before deleting the class.",
        )
        return HttpResponseRedirect(reverse_lazy("view_class", kwargs={"access_code": access_code}))

    klass.anonymise()

    return HttpResponseRedirect(reverse_lazy("dashboard") + "#classes")


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_delete_students(request, access_code):
    klass = get_object_or_404(Class, access_code=access_code)

    # check user is authorised to deal with class
    check_teacher_authorised(request, klass.teacher)

    # get student objects for students to be deleted, confirming they are in the class
    student_ids = json.loads(request.POST.get("transfer_students", "[]"))
    students = [get_object_or_404(Student, id=i, class_field=klass) for i in student_ids]

    def __anonymise(user):
        # Delete all personal data from inactive user and mark as inactive.
        # Student only has random username, password, first_name
        user.first_name = "Deleted"
        user.last_name = "User"
        user.is_active = False
        user.save()

    # Delete all of the students
    for student in students:
        # If the student has previously logged in, anonymise
        user = student.new_user
        if user.last_login:
            __anonymise(user)
            # remove login id so they can't log in with direct link anymore
            student.login_id = ""
            student.save()
        else:  # otherwise, just delete
            student.new_user.delete()

    return HttpResponseRedirect(reverse_lazy("view_class", kwargs={"access_code": access_code}))


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_edit_class(request, access_code):
    """
    Editing additional class details. Provides functionality for:
    - Editing the class name, sharing and joining settings
    - Locking or unlocking specific Rapid Router levels
    - Transferring the class to another teacher
    """
    klass = get_object_or_404(Class, access_code=access_code)
    old_teacher = klass.teacher
    other_teachers = Teacher.objects.filter(school=old_teacher.school).exclude(user=old_teacher.user)

    # check user authorised to see class
    check_teacher_authorised(request, klass.teacher)

    external_requests_message = klass.get_requests_message()

    blockly_episodes = get_blockly_episodes(request)
    python_episodes = get_python_episodes(request)

    locked_levels = klass.locked_levels.all()
    locked_levels_ids = [locked_level.id for locked_level in locked_levels]

    locked_worksheet_ids = [worksheet.id for worksheet in klass.locked_worksheets.all()]

    form = ClassEditForm(
        initial={
            "name": klass.name,
            "classmate_progress": klass.classmates_data_viewable,
        }
    )
    level_control_form = ClassLevelControlForm()
    class_move_form = ClassMoveForm(other_teachers)

    if request.method == "POST":
        if "class_edit_submit" in request.POST:
            form = ClassEditForm(request.POST)
            if form.is_valid():
                return process_edit_class_form(request, klass, form)
        elif "level_control_submit" in request.POST:
            level_control_form = ClassLevelControlForm(request.POST)
            if level_control_form.is_valid():
                return process_level_control_form(request, klass, blockly_episodes, python_episodes)
        elif "class_move_submit" in request.POST:
            class_move_form = ClassMoveForm(other_teachers, request.POST)
            if class_move_form.is_valid():
                return process_move_class_form(request, klass, class_move_form)

    return render(
        request,
        "portal/teach/teacher_edit_class.html",
        {
            "form": form,
            "class_move_form": class_move_form,
            "level_control_form": level_control_form,
            "blockly_episodes": blockly_episodes,
            "python_episodes": python_episodes,
            "locked_levels": locked_levels_ids,
            "locked_worksheet_ids": locked_worksheet_ids,
            "class": klass,
            "external_requests_message": external_requests_message,
        },
    )


def process_edit_class_form(request, klass, form):
    name = form.cleaned_data["name"]
    classmate_progress = bool(form.cleaned_data["classmate_progress"])

    external_requests_setting = form.cleaned_data["external_requests"]
    if external_requests_setting != "":
        # Change submitted for external requests
        hours = int(external_requests_setting)
        if hours == 0:
            # Setting to off
            klass.always_accept_requests = False
            klass.accept_requests_until = None
            messages.info(
                request,
                "Class set successfully to never receive requests from external students.",
            )

        elif hours < 1000:
            # Setting to number of hours
            klass.always_accept_requests = False
            klass.accept_requests_until = timezone.now() + timedelta(hours=hours)
            messages.info(
                request,
                "Class set successfully to receive requests from external students until "
                + klass.accept_requests_until.strftime("%d-%m-%Y %H:%M")
                + " "
                + timezone.get_current_timezone_name(),
            )

        else:
            # Setting to always on
            klass.always_accept_requests = True
            klass.accept_requests_until = None
            messages.info(
                request,
                "Class set successfully to always receive requests from external students (not recommended)",
            )

    klass.name = name
    klass.classmates_data_viewable = classmate_progress
    klass.save()

    messages.success(request, "The class's settings have been changed successfully.")

    return HttpResponseRedirect(reverse_lazy("view_class", kwargs={"access_code": klass.access_code}))


def process_level_control_form(request, klass: Class, blockly_episodes, python_episodes):
    """
    Find the levels that the user wants to lock and lock them for the specific class.
    :param request: The request sent by the user submitting the form.
    :param klass: The class for which the levels are being locked / unlocked.
    :param blockly_episodes: The set of Blockly Episodes (Rapid Router).
    :param blockly_episodes: The set of Python Episodes (Python Den).
    :return: A redirect to the teacher dashboard with a success message.
    """
    levels_to_lock_ids = []
    locked_worksheet_ids = []

    mark_levels_to_lock_in_episodes(request, blockly_episodes, levels_to_lock_ids, locked_worksheet_ids)
    mark_levels_to_lock_in_episodes(request, python_episodes, levels_to_lock_ids, locked_worksheet_ids)

    klass.locked_levels.clear()
    [klass.locked_levels.add(levels_to_lock_id) for levels_to_lock_id in levels_to_lock_ids]
    klass.locked_worksheets.clear()
    for locked_worksheet_id in locked_worksheet_ids:
        klass.locked_worksheets.add(locked_worksheet_id)

    messages.success(request, "Your level preferences have been saved.")
    activity_today = DailyActivity.objects.get_or_create(date=datetime.now().date())[0]
    activity_today.level_control_submits += 1
    activity_today.save()

    return HttpResponseRedirect(reverse_lazy("dashboard"))


def mark_levels_to_lock_in_episodes(request, episodes, levels_to_lock_ids, locked_worksheet_ids: list):
    """
    For a given set of Episodes, find which Levels are to be locked. This is done by checking the POST request data.
    If a Level ID is missing from the request.POST, it means it needs to be locked, and if the entire Episode is missing
    from the request.POST, it means all the Levels under that Episode need to be locked.
    :param request: The request sent by the user submitting the form.
    :param episodes: The set of Episodes, in this case either the Blockly Episodes or the Python Episodes.
    :param levels_to_lock_ids: A list of Level IDs marked to be locked.
    """
    for episode in episodes:
        episode_levels = episode["levels"]
        episode_worksheets = episode["worksheets"]
        episode_index = f"episode{episode['id']}"
        if episode_index in request.POST:
            [
                levels_to_lock_ids.append(episode_level["id"])
                for episode_level in episode_levels
                if f'level:{episode_level["id"]}' not in request.POST.getlist(episode_index)
            ]
            for episode_worksheet in episode_worksheets:
                worksheet_id = episode_worksheet["id"]
                if f"worksheet:{worksheet_id}" not in request.POST.getlist(episode_index):
                    locked_worksheet_ids.append(worksheet_id)
        else:
            [levels_to_lock_ids.append(episode_level["id"]) for episode_level in episode_levels]
            for episode_worksheet in episode_worksheets:
                locked_worksheet_ids.append(episode_worksheet["id"])


def process_move_class_form(request, klass, form):
    new_teacher_id = form.cleaned_data["new_teacher"]
    new_teacher = get_object_or_404(Teacher, id=new_teacher_id)

    klass.teacher = new_teacher
    klass.save()

    messages.success(
        request,
        "The class has been successfully assigned to a different teacher.",
    )
    return HttpResponseRedirect(reverse_lazy("dashboard"))


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_edit_student(request, pk):
    """
    Changing a student's details
    """
    student = get_object_or_404(Student, id=pk)
    check_teacher_authorised(request, student.class_field.teacher)

    name_form = TeacherEditStudentForm(student, initial={"name": student.new_user.first_name})

    password_form = TeacherSetStudentPass()
    set_password_mode = False

    if request.method == "POST":
        if "update_details" in request.POST:
            name_form = TeacherEditStudentForm(student, request.POST)
            if name_form.is_valid():
                name = name_form.cleaned_data["name"]
                student.new_user.first_name = name
                student.new_user.save()
                student.save()

                messages.success(
                    request,
                    "The student's details have been changed successfully.",
                )

                return HttpResponseRedirect(
                    reverse_lazy(
                        "view_class",
                        kwargs={"access_code": student.class_field.access_code},
                    )
                )

        else:
            password_form = TeacherSetStudentPass(request.POST)
            if password_form.is_valid():
                return process_reset_password_form(request, student, password_form)
            set_password_mode = True

    return render(
        request,
        "portal/teach/teacher_edit_student.html",
        {
            "name_form": name_form,
            "password_form": password_form,
            "student": student,
            "class": student.class_field,
            "set_password_mode": set_password_mode,
        },
    )


def process_reset_password_form(request, student, password_form):
    # check not default value for CharField
    new_password = password_form.cleaned_data["password"]
    if new_password:
        # generate uuid for url and store the hashed
        uuidstr = uuid4().hex
        login_id = get_hashed_login_id(uuidstr)
        login_url = request.build_absolute_uri(
            reverse(
                "student_direct_login",
                kwargs={"user_id": student.new_user.id, "login_id": uuidstr},
            )
        )

        students_info = [
            {
                "id": student.new_user.id,
                "name": student.new_user.first_name,
                "password": new_password,
                "login_url": login_url,
            }
        ]

        handle_reset_password_tracking(request, "SCHOOL_STUDENT")
        student.new_user.set_password(new_password)
        student.new_user.save()
        student.login_id = login_id
        clear_ratelimit_cache_for_user(f"{student.new_user.first_name},{student.class_field.access_code}")
        student.blocked_time = datetime.now(tz=pytz.utc) - timedelta(days=1)
        student.save()

        return render(
            request,
            "portal/teach/onboarding_print.html",
            {
                "class": student.class_field,
                "students_info": students_info,
                "onboarding_done": True,
                "query_data": json.dumps(students_info),
                "class_url": request.build_absolute_uri(
                    reverse(
                        "student_login",
                        kwargs={"access_code": student.class_field.access_code},
                    )
                ),
            },
        )


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_dismiss_students(request, access_code):
    """
    Dismiss a student (make them independent)
    """
    klass = get_object_or_404(Class, access_code=access_code)

    check_teacher_authorised(request, klass.teacher)

    # get student objects for students to be dismissed, confirming they are in the class
    student_ids = json.loads(request.POST.get("transfer_students", "[]"))
    students = [get_object_or_404(Student, id=i, class_field=klass) for i in student_ids]

    TeacherDismissStudentsFormSet = formset_factory(
        wraps(TeacherDismissStudentsForm)(partial(TeacherDismissStudentsForm)),
        extra=0,
        formset=BaseTeacherDismissStudentsFormSet,
    )

    if is_right_dismiss_form(request):
        formset = TeacherDismissStudentsFormSet(request.POST)
        if formset.is_valid():
            return process_dismiss_student_form(request, formset, klass, access_code)

    else:
        initial_data = [
            {
                "orig_name": student.new_user.first_name,
                "name": student.new_user.first_name,
                "email": "",
            }
            for student in students
        ]

        formset = TeacherDismissStudentsFormSet(initial=initial_data)

    return render(
        request,
        "portal/teach/teacher_dismiss_students.html",
        {"formset": formset, "class": klass, "students": students},
    )


def is_right_dismiss_form(request):
    return request.method == "POST" and "submit_dismiss" in request.POST


def process_dismiss_student_form(request, formset, klass, access_code):
    failed_users = []  # users that failed to be transferred
    for data in formset.cleaned_data:
        # check if email is already used
        users_with_email = User.objects.filter(email=data["email"])
        # email is already taken, skip this user
        if users_with_email.exists():
            failed_users.append(data["orig_name"])
            continue

        student = get_object_or_404(
            Student,
            class_field=klass,
            new_user__first_name__iexact=data["orig_name"],
        )

        students_levels = Level.objects.filter(owner=student.new_user.userprofile).all()
        for level in students_levels:
            level.shared_with.set([])
            level.save()

        student.class_field = None
        student.new_user.first_name = data["name"]
        student.new_user.username = data["email"]
        student.new_user.email = data["email"]
        student.user.is_verified = False
        student.save()
        student.new_user.save()
        student.user.save()

        # log the data
        joinrelease = JoinReleaseStudent.objects.create(student=student, action_type=JoinReleaseStudent.RELEASE)
        joinrelease.save()

        send_verification_email(request, student.new_user, data, school=klass.teacher.school)

    if not failed_users:
        messages.success(
            request,
            "The students have been released successfully from the class.",
        )
    else:
        messages.warning(
            request,
            f"The following students could not be released: {', '.join(failed_users)}. "
            "Please make sure the email has not been registered to another account.",
        )

    return HttpResponseRedirect(reverse_lazy("view_class", kwargs={"access_code": access_code}))


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_class_password_reset(request, access_code):
    """
    Reset passwords for one or more students
    """
    klass = get_object_or_404(Class, access_code=access_code)

    # check user authorised to see class
    check_teacher_authorised(request, klass.teacher)

    student_ids = json.loads(request.POST.get("transfer_students", "[]"))
    students = [get_object_or_404(Student, id=i, class_field=klass) for i in student_ids]

    students_info = []
    handle_reset_password_tracking(request, "SCHOOL_STUDENT", access_code)
    for student in students:
        password = generate_password(STUDENT_PASSWORD_LENGTH)

        # generate uuid for url and store the hashed
        login_id, hashed_login_id = generate_login_id()
        login_url = generate_student_url(request, student, login_id)

        students_info.append(
            {
                "id": student.new_user.id,
                "name": student.new_user.first_name,
                "password": password,
                "login_url": login_url,
            }
        )
        student.new_user.set_password(password)
        student.new_user.save()
        student.login_id = hashed_login_id
        clear_ratelimit_cache_for_user(f"{student.new_user.first_name},{access_code}")
        student.blocked_time = datetime.now(tz=pytz.utc) - timedelta(days=1)
        student.save()

    return render(
        request,
        "portal/teach/onboarding_print.html",
        {
            "class": klass,
            "onboarding_done": True,
            "passwords_reset": True,
            "students_info": students_info,
            "query_data": json.dumps(students_info),
            "class_url": request.build_absolute_uri(
                reverse("student_login", kwargs={"access_code": klass.access_code})
            ),
        },
    )


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_move_students(request, access_code):
    """
    Move students
    """
    klass = get_object_or_404(Class, access_code=access_code)

    # check user is authorised to deal with class
    check_teacher_authorised(request, klass.teacher)

    transfer_students = request.POST.get("transfer_students", "[]")

    school = klass.teacher.school

    # get classes in same school
    classes = school.classes()
    classes.remove(klass)

    form = TeacherMoveStudentsDestinationForm(classes)

    return render(
        request,
        "portal/teach/teacher_move_students.html",
        {
            "transfer_students": transfer_students,
            "old_class": klass,
            "form": form,
        },
    )


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_move_students_to_class(request, access_code):
    """
    Disambiguation for moving students (teacher gets to rename the students to avoid clashes)
    """
    old_class = get_object_or_404(Class, access_code=access_code)
    new_class_id = request.POST.get("new_class", None)
    new_class = get_object_or_404(Class, id=new_class_id)

    check_if_move_authorised(request, old_class, new_class)

    transfer_students_ids = json.loads(request.POST.get("transfer_students", "[]"))

    # get student objects for students to be transferred, confirming they are in the old class still
    transfer_students = [get_object_or_404(Student, id=i, class_field=old_class) for i in transfer_students_ids]

    # get new class' students
    new_class_students = Student.objects.filter(class_field=new_class, new_user__is_active=True).order_by(
        "new_user__first_name"
    )

    TeacherMoveStudentDisambiguationFormSet = formset_factory(
        wraps(TeacherMoveStudentDisambiguationForm)(partial(TeacherMoveStudentDisambiguationForm)),
        extra=0,
        formset=BaseTeacherMoveStudentsDisambiguationFormSet,
    )

    if is_right_move_form(request):
        formset = TeacherMoveStudentDisambiguationFormSet(new_class, request.POST)
        if formset.is_valid():
            return process_move_students_form(request, formset, old_class, new_class)
    else:
        # format the students for the form
        initial_data = [
            {
                "orig_name": student.new_user.first_name,
                "name": student.new_user.first_name,
            }
            for student in transfer_students
        ]

        formset = TeacherMoveStudentDisambiguationFormSet(new_class, initial=initial_data)

    return render(
        request,
        "portal/teach/teacher_move_students_to_class.html",
        {
            "formset": formset,
            "old_class": old_class,
            "new_class": new_class,
            "new_class_students": new_class_students,
            "transfer_students": transfer_students,
        },
    )


def check_if_move_authorised(request, old_class, new_class):
    teacher = request.user.new_teacher

    # check teacher has permission to edit old_class and that both classes
    # are in the same school
    if (not teacher.is_admin and teacher != old_class.teacher) or teacher.school != new_class.teacher.school:
        raise Http404


def is_right_move_form(request):
    return request.method == "POST" and "submit_disambiguation" in request.POST


def process_move_students_form(request, formset, old_class, new_class):
    old_teacher = old_class.teacher
    new_teacher = new_class.teacher

    for name_update in formset.cleaned_data:
        student = get_object_or_404(
            Student,
            class_field=old_class,
            new_user__first_name__iexact=name_update["orig_name"],
        )
        student.class_field = new_class
        student.new_user.first_name = name_update["name"]

        student.save()
        student.new_user.save()

    messages.success(request, "The students have been transferred successfully.")
    return HttpResponseRedirect(reverse_lazy("view_class", kwargs={"access_code": old_class.access_code}))


class DownloadType(Enum):
    CSV = 1
    LOGIN_CARDS = 2
    PRIMARY_PACK = 3
    PYTHON_PACK = 4


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_print_reminder_cards(request, access_code):
    response = HttpResponse(content_type="application/pdf")
    response["Content-Disposition"] = 'filename="student_reminder_cards.pdf"'

    p = canvas.Canvas(response, pagesize=A4)

    # Define constants that determine the look of the cards
    PAGE_WIDTH, PAGE_HEIGHT = A4
    PAGE_MARGIN = PAGE_WIDTH // 16
    INTER_CARD_MARGIN = PAGE_WIDTH // 64
    CARD_PADDING = PAGE_WIDTH // 48

    # rows and columns on page
    NUM_X = REMINDER_CARDS_PDF_COLUMNS
    NUM_Y = REMINDER_CARDS_PDF_ROWS

    CARD_WIDTH = (PAGE_WIDTH - PAGE_MARGIN * 2) // NUM_X
    CARD_HEIGHT = (PAGE_HEIGHT - PAGE_MARGIN * 4) // NUM_Y

    CARD_INNER_HEIGHT = CARD_HEIGHT - CARD_PADDING * 2

    logo_image = ImageReader(staticfiles_storage.path("portal/img/logo_cfl_reminder_cards.jpg"))

    klass = get_object_or_404(Class, access_code=access_code)
    # Check auth
    check_teacher_authorised(request, klass.teacher)

    # Use data from the query string if given
    student_data = get_student_data(request)
    student_login_link = request.build_absolute_uri(reverse("student_login_access_code"))
    class_login_link = request.build_absolute_uri(reverse("student_login", kwargs={"access_code": access_code}))

    # Now draw everything
    x = 0
    y = 0

    current_student_count = 0
    for student in student_data:
        # warning text for every new page
        if current_student_count % (NUM_X * NUM_Y) == 0:
            p.setFillColor(red)
            p.setFont("Helvetica-Bold", 10)
            p.drawString(PAGE_MARGIN, PAGE_MARGIN / 2, REMINDER_CARDS_PDF_WARNING_TEXT)

        left = PAGE_MARGIN + x * CARD_WIDTH + x * INTER_CARD_MARGIN * 2
        bottom = PAGE_HEIGHT - PAGE_MARGIN - (y + 1) * CARD_HEIGHT - y * INTER_CARD_MARGIN

        inner_bottom = bottom + CARD_PADDING

        # card border
        p.setStrokeColor(black)
        p.rect(left, bottom, CARD_WIDTH, CARD_HEIGHT)

        card_logo_height = CARD_HEIGHT - INTER_CARD_MARGIN * 2

        # logo
        p.drawImage(
            logo_image,
            left + INTER_CARD_MARGIN,
            bottom + INTER_CARD_MARGIN,
            height=card_logo_height,
            preserveAspectRatio=True,
            anchor="w",
        )

        text_left = left + INTER_CARD_MARGIN + (logo_image.getSize()[0] / logo_image.getSize()[1]) * card_logo_height

        # student details
        p.setFillColor(black)
        p.setFont("Helvetica", 12)
        p.drawString(
            text_left,
            inner_bottom + CARD_INNER_HEIGHT * 0.9,
            f"Class code: {klass.access_code} at {student_login_link}",
        )
        p.setFont("Helvetica-BoldOblique", 12)
        p.drawString(text_left, inner_bottom + CARD_INNER_HEIGHT * 0.6, "OR")
        p.setFont("Helvetica", 12)
        p.drawString(
            text_left + 22,
            inner_bottom + CARD_INNER_HEIGHT * 0.6,
            f"class link: {class_login_link}",
        )
        p.drawString(
            text_left,
            inner_bottom + CARD_INNER_HEIGHT * 0.3,
            f"Name: {student['name']}",
        )
        p.drawString(text_left, inner_bottom, f"Password: {student['password']}")

        x = (x + 1) % NUM_X
        y = compute_show_page_character(p, x, y, NUM_Y)
        current_student_count += 1

    compute_show_page_end(p, x, y)

    p.save()

    count_student_details_click(DownloadType.LOGIN_CARDS)

    return response


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_download_csv(request, access_code):
    response = HttpResponse(content_type="text/csv")
    response["Content-Disposition"] = 'attachment; filename="student_login_urls.csv"'

    klass = get_object_or_404(Class, access_code=access_code)
    # Check auth
    check_teacher_authorised(request, klass.teacher)

    class_url = request.build_absolute_uri(reverse("student_login", kwargs={"access_code": access_code}))

    # Use data from the query string if given
    student_data = get_student_data(request)
    if student_data:
        writer = csv.writer(response)
        writer.writerow([access_code, class_url])
        for student in student_data:
            writer.writerow([student["name"], student["password"], student["login_url"]])

    count_student_details_click(DownloadType.CSV)

    return response


def get_student_data(request):
    if request.method == "POST":
        data = request.POST.get("data", "[]")
        return json.loads(data)
    return []


def compute_show_page_character(p, x, y, NUM_Y):
    if x == 0:
        y = (y + 1) % NUM_Y
        if y == 0:
            p.showPage()
    return y


def compute_show_page_end(p, x, y):
    if x != 0 or y != 0:
        p.showPage()


def count_student_pack_downloads_click(student_pack_type):
    activity_today = DailyActivity.objects.get_or_create(date=datetime.now().date())[0]
    if DownloadType(student_pack_type) == DownloadType.PRIMARY_PACK:
        activity_today.primary_coding_club_downloads += 1
    else:
        raise Exception("Unknown download type")
    activity_today.save()


def count_student_details_click(download_type):
    activity_today = DailyActivity.objects.get_or_create(date=datetime.now().date())[0]

    if download_type == DownloadType.CSV:
        activity_today.csv_click_count += 1
    elif download_type == DownloadType.LOGIN_CARDS:
        activity_today.login_cards_click_count += 1
    else:
        raise Exception("Unknown download type")

    activity_today.save()



from common.mail import campaign_ids, send_dotdigital_email
from common.models import Student, Teacher
from django import forms
from django.contrib.auth import forms as django_auth_forms
from django.contrib.auth import get_user_model
from django.contrib.auth.tokens import default_token_generator
from django.contrib.sites.shortcuts import get_current_site
from django.urls import reverse_lazy
from django.utils.encoding import force_bytes
from django.utils.http import urlsafe_base64_encode
from django_recaptcha.fields import ReCaptchaField
from django_recaptcha.widgets import ReCaptchaV2Invisible

from portal.helpers.password import PasswordStrength, form_clean_password


class TeacherPasswordResetSetPasswordForm(django_auth_forms.SetPasswordForm):
    def __init__(self, user, *args, **kwargs):
        super(TeacherPasswordResetSetPasswordForm, self).__init__(user, *args, **kwargs)
        self.fields["new_password1"].help_text = "Enter your new password"
        self.fields["new_password1"].widget.attrs["placeholder"] = "New password"
        self.fields["new_password1"].widget.attrs["autocomplete"] = "off"
        self.fields["new_password2"].help_text = "Confirm your new password"
        self.fields["new_password2"].widget.attrs["placeholder"] = "Confirm password"
        self.fields["new_password2"].widget.attrs["autocomplete"] = "off"

    def clean_new_password1(self):
        return form_clean_password(self, "new_password1", PasswordStrength.TEACHER)


class StudentPasswordResetSetPasswordForm(django_auth_forms.SetPasswordForm):
    def __init__(self, user, *args, **kwargs):
        super(StudentPasswordResetSetPasswordForm, self).__init__(user, *args, **kwargs)
        self.fields["new_password1"].help_text = "Enter your new password"
        self.fields["new_password1"].widget.attrs["placeholder"] = "New password"
        self.fields["new_password1"].widget.attrs["autocomplete"] = "off"
        self.fields["new_password2"].help_text = "Confirm your new password"
        self.fields["new_password2"].widget.attrs["placeholder"] = "Confirm password"
        self.fields["new_password2"].widget.attrs["autocomplete"] = "off"

    def clean_new_password1(self):
        return form_clean_password(self, "new_password1", PasswordStrength.INDEPENDENT)


class PasswordResetForm(forms.Form):
    email = forms.EmailField(
        widget=forms.EmailInput(attrs={"autocomplete": "off", "placeholder": "Email address"}),
        help_text="Enter your email address",
    )

    captcha = ReCaptchaField(widget=ReCaptchaV2Invisible)

    def save(
        self,
        domain_override=None,
        subject_template_name=None,
        email_template_name="portal/reset_password_email.txt",
        use_https=False,
        token_generator=default_token_generator,
        from_email=None,
        request=None,
        html_email_template_name="portal/reset_password_email.html",
    ):
        """
        Generates a one-use only link for resetting password and sends to the
        user.
        """
        UserModel = get_user_model()
        if self.username:
            active_users = UserModel._default_manager.filter(username=self.username, is_active=True)
            for user in active_users:
                # Make sure that no email is sent to a user that actually has
                # a password marked as unusable
                if not user.has_usable_password():
                    continue
                if not domain_override:
                    current_site = get_current_site(request)
                    domain = current_site.domain
                else:
                    domain = domain_override

                reset_password_uri = reverse_lazy(
                    "password_reset_check_and_confirm",
                    kwargs={
                        "uidb64": urlsafe_base64_encode(force_bytes(user.pk)),
                        "token": token_generator.make_token(user),
                    },
                )
                protocol = self._compute_protocol(use_https)
                reset_password_url = f"{protocol}://{domain}{reset_password_uri}"

                send_dotdigital_email(
                    campaign_ids["reset_password"],
                    [user.email],
                    personalization_values={"RESET_PASSWORD_LINK": reset_password_url},
                )

    def _compute_protocol(self, use_https):
        return "https" if use_https else "http"


class TeacherPasswordResetForm(PasswordResetForm):
    def clean_email(self):
        email = self.cleaned_data.get("email", None)
        self.username = ""
        teacher = Teacher.objects.filter(new_user__email=email)
        # Check such an email exists
        if teacher.exists():
            self.username = teacher[0].new_user.username
        return email


class StudentPasswordResetForm(PasswordResetForm):
    def clean_email(self):
        email = self.cleaned_data.get("email", None)
        self.username = ""
        student = Student.objects.filter(new_user__email=email)
        # Check such an email exists
        if student.exists():
            self.username = student[0].new_user.username
        return email


class DeleteAccountForm(forms.Form):
    delete_password = forms.CharField(
        required=True,
        widget=forms.PasswordInput(attrs={"autocomplete": "off", "placeholder": "Confirm password"}),
        help_text="Confirm password",
    )

    unsubscribe_newsletter = forms.BooleanField(
        label="Please remove me from the newsletter and marketing emails too.",
        widget=forms.CheckboxInput(),
        initial=True,
        required=False,
    )

    def __init__(self, user, *args, **kwargs):
        self.user = user
        super(DeleteAccountForm, self).__init__(*args, **kwargs)

    def clean(self):
        delete_password = self.cleaned_data.get("delete_password", None)
        if not self.user.check_password(delete_password):
            raise forms.ValidationError("Incorrect password")



# Generated by Django 1.10.8 on 2019-08-27 08:37
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("portal", "0055_add_preview_user")]

    operations = [
        migrations.RemoveField(model_name="userprofile", name="preview_user"),
        migrations.RemoveField(model_name="school", name="eligible_for_testing"),
    ]



# Generated by Django 1.11.24 on 2020-03-02 13:10
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("portal", "0056_remove_preview_user")]

    operations = [migrations.DeleteModel(name="FrontPageNews")]



from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):
    dependencies = [("portal", "0049_refactor_emailverifications")]

    operations = [
        migrations.RemoveField(model_name="emailverification", name="user"),
        migrations.RemoveField(model_name="emailverification", name="used"),
        migrations.AddField(
            model_name="emailverification",
            name="user",
            field=models.ForeignKey(
                to=settings.AUTH_USER_MODEL,
                related_name="email_verifications",
                null=True,
                blank=True,
                on_delete=models.CASCADE,
            ),
        ),
        migrations.RunSQL(
            ("UPDATE portal_emailverification" "  SET user_id = new_user_id;")
        ),
    ]



import pytest


@pytest.mark.django_db
def test_front_page_news_model_removed(migrator):
    migrator.apply_initial_migration(("portal", "0056_remove_preview_user"))
    new_state = migrator.apply_tested_migration(("portal", "0057_delete_frontpagenews"))

    model_names = [model._meta.db_table for model in new_state.apps.get_models()]
    assert "portal_frontpagenews" not in model_names



from selenium.webdriver.common.by import By

from .base_page import BasePage
from .password_reset_page import PasswordResetPage
from .play import dashboard_page


class IndependentStudentLoginPage(BasePage):
    def __init__(self, browser):
        super(IndependentStudentLoginPage, self).__init__(browser)

        assert self.on_correct_page("independent_login_page")

    def independent_student_login(self, email, password):
        self._independent_student_login(email, password)

        return dashboard_page.PlayDashboardPage(self.browser)

    def independent_student_login_failure(self, email, password):
        self._independent_student_login(email, password)
        return self

    def _independent_student_login(self, email, password):
        self.browser.find_element(By.ID, "id_username").send_keys(email)
        self.browser.find_element(By.ID, "id_password").send_keys(password)
        self.browser.find_element(By.NAME, "independent_student_login").click()

    def go_to_indep_forgotten_password_page(self):
        self.browser.find_element(By.ID, "student_forgotten_password_button").click()
        return PasswordResetPage(self.browser)

    def has_login_failed(self, form_id, error):
        errors = self.browser.find_element(By.ID, form_id).find_element(By.CLASS_NAME, "errorlist").text
        return error in errors



from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [("portal", "0047_remove_userprofile_avatar")]

    operations = [
        migrations.AlterModelOptions(
            name="frontpagenews", options={"verbose_name_plural": "front page news"}
        )
    ]



from selenium.webdriver.common.by import By

from portal.tests.pageObjects.portal.teach.class_page import TeachClassPage
from portal.tests.pageObjects.portal.teach.teach_base_page import TeachBasePage


class AddedIndependentStudentToClassPage(TeachBasePage):
    def __init__(self, browser):
        super(AddedIndependentStudentToClassPage, self).__init__(browser)

        assert self.on_correct_page("added_independent_student_to_class")

    def return_to_class(self):
        self.browser.find_element(By.ID, "return_button").click()

        return TeachClassPage(self.browser)



from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings


class Migration(migrations.Migration):
    dependencies = [("portal", "0052_refactor_emailverifications_3")]

    operations = [
        migrations.AddField(
            model_name="teacher",
            name="new_user",
            field=models.OneToOneField(
                to=settings.AUTH_USER_MODEL,
                related_name="new_teacher",
                null=True,
                blank=True,
                on_delete=models.CASCADE,
            ),
        ),
        migrations.AddField(
            model_name="student",
            name="new_user",
            field=models.OneToOneField(
                to=settings.AUTH_USER_MODEL,
                related_name="new_student",
                null=True,
                blank=True,
                on_delete=models.CASCADE,
            ),
        ),
        migrations.AddField(
            model_name="guardian",
            name="new_user",
            field=models.OneToOneField(
                to=settings.AUTH_USER_MODEL,
                related_name="new_guardian",
                null=True,
                blank=True,
                on_delete=models.CASCADE,
            ),
        ),
        migrations.RunSQL(
            (
                "UPDATE portal_teacher SET new_user_id = ("
                "  SELECT portal_userprofile.user_id FROM portal_userprofile "
                "    WHERE portal_userprofile.id = portal_teacher.user_id);"
            )
        ),
        migrations.RunSQL(
            (
                "UPDATE portal_student SET new_user_id = ("
                "  SELECT portal_userprofile.user_id FROM portal_userprofile "
                "    WHERE portal_userprofile.id = portal_student.user_id);"
            )
        ),
        migrations.RunSQL(
            (
                "UPDATE portal_guardian SET new_user_id = ("
                "  SELECT portal_userprofile.user_id FROM portal_userprofile "
                "    WHERE portal_userprofile.id = portal_guardian.user_id);"
            )
        ),
    ]



from portal.helpers.regexes import ACCESS_CODE_FROM_URL_PATTERN


def get_access_code_from_request(request):
    try:
        access_code = ACCESS_CODE_FROM_URL_PATTERN.search(request.get_full_path()).group(1)
    except AttributeError:
        access_code = ""
        print(f"Access code not found in {request.get_full_path()}")
    return access_code



from __future__ import unicode_literals

from django.db import models, migrations
import django_countries.fields


class Migration(migrations.Migration):

    dependencies = [("portal", "0044_auto_20150430_0959")]

    operations = [
        migrations.AlterField(
            model_name="school",
            name="country",
            field=django_countries.fields.CountryField(max_length=2),
        )
    ]



from builtins import map


def is_captcha_in_form(form):
    return "captcha" in form.fields


def remove_captcha_from_forms(*args):
    list(map(remove_captcha_from_form, args))


def remove_captcha_from_form(form):
    form.fields.pop("captcha", None)



"""Libsass compiler for django-pipeline.
Speedups development and/or production when compiling sass assets. No need of
ruby sass anymore.
"""

import sass
import codecs
from pipeline.compilers import CompilerBase
from django.conf import settings


class LibSassCompiler(CompilerBase):
    output_extension = "css"

    def match_file(self, filename):
        return filename.endswith((".scss", ".sass"))

    def compile_file(self, infile, outfile, outdated=False, force=False):
        myfile = codecs.open(outfile, "w", "utf-8")

        if settings.DEBUG:
            myfile.write(sass.compile(filename=infile))
        else:
            myfile.write(sass.compile(filename=infile, output_style="compressed"))
        return myfile.close()



def has_beta_access(request):
    return is_developer(request)


def is_on_beta_host(request):
    return request.get_host().startswith("beta")


def is_developer(request):
    return (not request.user.is_anonymous) and request.user.userprofile.developer



from django.middleware.security import SecurityMiddleware


class CustomSecurityMiddleware(SecurityMiddleware):
    """
    Extends Django's Security Middleware.
    See https://docs.djangoproject.com/en/4.2/_modules/django/middleware/security/ for
    the source code, as well as https://docs.djangoproject.com/en/4.2/ref/middleware/#module-django.middleware.security
    for docs on security middleware.
    """

    def process_response(self, request, response):
        """
        Extends the original security middleware to ensure the X-XSS-Protection header
        is set to 1.
        https://docs.djangoproject.com/en/5.1/releases/4.0/#securitymiddleware-no-longer-sets-the-x-xss-protection-header
        """
        super().process_response(request, response)
        response.headers.setdefault("X-XSS-Protection", "1; mode=block")

        return response



HOME_LEARNING_BANNER = {
    "title": "Home learning",
    "subtitle": "Whether you're a parent, a caregiver, or a curious student — our Rapid Router game is easy to use "
    "and free – forever.",
    "text": "",
    "image_class": "banner--picture--homelearning",
}



import time
from typing import Tuple
from unittest import mock

from _pytest.monkeypatch import MonkeyPatch
from common.tests.utils.classes import create_class_directly
from common.tests.utils.organisation import create_organisation_directly
from common.tests.utils.student import create_school_student_directly
from common.tests.utils.teacher import signup_teacher_directly
from django.contrib import auth
from django.contrib.auth.models import User
from django.http import HttpResponseRedirect
from django.test import Client, TestCase
from django.urls import reverse

MOCKED_SESSION_EXPIRY_TIME = 5


class TestAdminAccessMiddleware(TestCase):
    """
    This tests the AdminAccessMiddleware class by checking that users are redirected
    to the correct pages depending on their permissions, upon request to access the
    admin pages. Specifically:
    - An unauthenticated user should be redirected to the teacher login.
    - An authenticated user who is a superuser, OR has 2FA enabled, or neither, is
    redirected to the teacher dashboard.
    - An authenticated user who is a superuser AND has 2FA enabled isn't redirected.
    """

    def setUp(self) -> None:
        self.client = Client()
        self.email, self.password = self._setup_user()

        self.monkeypatch = MonkeyPatch()
        self.monkeypatch.setattr(
            "deploy.middleware.admin_access.MODULE_NAME", "test"
        )

    def _setup_user(self) -> Tuple[str, str]:
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        return email, password

    def _make_user_superuser(self) -> None:
        user = User.objects.get(email=self.email)
        user.is_superuser = True
        user.is_staff = True
        user.save()

    def test_unauthenticated_user_is_redirected(self):
        response = self.client.get("/administration/")

        assert response.status_code == 302
        assert type(response) == HttpResponseRedirect
        assert response.url == "/login/teacher/"

    def test_authenticated_user_with_no_permissions_is_redirected(self):
        self.client.login(username=self.email, password=self.password)

        response = self.client.get("/administration/")

        self.client.logout()

        assert response.status_code == 302
        assert type(response) == HttpResponseRedirect
        assert response.url == "/teach/dashboard/"

    def test_superuser_without_2FA_is_redirected(self):
        self._make_user_superuser()

        self.client.login(username=self.email, password=self.password)

        response = self.client.get("/administration/")

        self.client.logout()

        assert response.status_code == 302
        assert type(response) == HttpResponseRedirect
        assert response.url == "/teach/dashboard/"

    @mock.patch(
        "deploy.middleware.admin_access.using_two_factor",
        return_value=True,
        autospec=True,
    )
    def test_non_superuser_with_2FA_is_redirected(self, mock_using_two_factor):
        self.client.login(username=self.email, password=self.password)

        response = self.client.get("/administration/")

        self.client.logout()

        assert response.status_code == 302
        assert type(response) == HttpResponseRedirect
        assert response.url == "/teach/dashboard/"

    @mock.patch(
        "deploy.middleware.admin_access.using_two_factor",
        return_value=True,
        autospec=True,
    )
    def test_superuser_with_2FA_can_access_admin_site(
        self, mock_using_two_factor
    ):
        self._make_user_superuser()

        self.client.login(username=self.email, password=self.password)

        response = self.client.get("/administration/")

        self.client.logout()

        assert response.status_code == 200


class TestSecurityMiddleware(TestCase):
    """
    This tests the SecurityMiddleware class and checks the correct security headers
    are in place.
    """

    def test_security_headers(self):
        client = Client()
        response = client.get("/")

        assert response.status_code == 200
        assert response.headers["cache-control"] == "private"
        assert response.headers["x-content-type-options"] == "nosniff"
        assert response.headers["x-frame-options"] == "DENY"
        assert response.headers["x-xss-protection"] == "1; mode=block"


class TestSessionTimeoutMiddleware(TestCase):
    """
    This tests the SessionTimeoutMiddleware class and checks the user is logged out
    after the defined time of inactivity.
    """

    def setUp(self) -> None:
        self.client = Client()
        self.email, self.password = self._setup_user()

        self.monkeypatch = MonkeyPatch()
        self.monkeypatch.setattr(
            "deploy.middleware.session_timeout.SESSION_EXPIRY_TIME", 5
        )

    def _setup_user(self) -> Tuple[str, str]:
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        return email, password

    def test_session_timeout(self):
        self.client.login(username=self.email, password=self.password)

        self.client.get("/")
        user = auth.get_user(self.client)

        assert user.is_authenticated

        time.sleep(MOCKED_SESSION_EXPIRY_TIME)

        response = self.client.get("/")
        user = auth.get_user(self.client)

        assert not user.is_authenticated

        messages = list(response.context["messages"])
        assert len(messages) > 0
        assert str(messages[0]) == "You have been logged out due to inactivity."

    def test_session_reset(self):
        self.client.login(username=self.email, password=self.password)

        self.client.get("/")
        user = auth.get_user(self.client)

        assert user.is_authenticated

        time.sleep(MOCKED_SESSION_EXPIRY_TIME * 0.66)

        url = reverse("reset_session_time")
        self.client.get(url)

        time.sleep(MOCKED_SESSION_EXPIRY_TIME * 0.66)

        self.client.get("/")
        user = auth.get_user(self.client)

        assert user.is_authenticated


class TestScreentimeWarningMiddleware(TestCase):
    """
    This tests the ScreentimeWarningMiddleware class and popup timeout is set properly.
    """

    def setUp(self) -> None:
        self.client = Client()
        self.email, self.password = self._setup_user()

    def _setup_user(self) -> Tuple[str, str]:
        email, password = signup_teacher_directly()
        create_organisation_directly(email)
        _, _, access_code = create_class_directly(email)
        create_school_student_directly(access_code)

        return email, password

    def test_screentime_warning_timeout(self):
        # Timeout should not be there if the user is not logged in
        session = self.client.session
        assert "screentime_warning_timeout" not in session

        # Log in as a teacher
        self.client.login(username=self.email, password=self.password)

        # Check the screentime_warning_timeout decreases after consecutive requests
        self.client.get("/")
        session = self.client.session
        assert "screentime_warning_timeout" in session
        previous_screentime_warning_timeout = session[
            "screentime_warning_timeout"
        ]

        self.client.get("/")
        session = self.client.session
        assert "screentime_warning_timeout" in session
        new_screentime_warning_timeout = session["screentime_warning_timeout"]

        assert (
            new_screentime_warning_timeout < previous_screentime_warning_timeout
        )

        # Check the reset_screentime_warning API resets the timeout
        url = reverse("reset_screentime_warning")
        self.client.get(url)
        self.client.get("/")
        session = self.client.session
        assert "screentime_warning_timeout" in session
        renewed_screentime_warning_timeout = session[
            "screentime_warning_timeout"
        ]
        assert (
            renewed_screentime_warning_timeout > new_screentime_warning_timeout
        )



from common.utils import two_factor_cache_key
from django.core.cache import cache
from django.db.models.signals import post_save, pre_delete
from django.dispatch import receiver
from django_otp.models import Device


@receiver([post_save, pre_delete])
def clear_two_factor_cache(sender, **kwargs):
    if issubclass(sender, Device):
        user = kwargs["instance"].user
        cache.delete(two_factor_cache_key(user))



from __future__ import absolute_import

from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import Select

from . import class_page
from . import dashboard_page
from .teach_base_page import TeachBasePage


class TeachMoveClassPage(TeachBasePage):
    def __init__(self, browser):
        super(TeachMoveClassPage, self).__init__(browser)

        assert self.on_correct_page("teach_move_class_page")

    def get_list_length(self):
        return len(self.browser.find_element(By.ID, "id_new_teacher").find_elements(By.TAG_NAME, "option"))

    def select_teacher_by_index(self, teacher_index):
        Select(self.browser.find_element(By.ID, "id_new_teacher")).select_by_index(teacher_index)
        return self

    def cancel(self):
        self.browser.find_element(By.ID, "cancel_button").click()
        return class_page.TeachClassPage(self.browser)

    def move(self):
        self.browser.find_element(By.ID, "move_button").click()
        return dashboard_page.TeachDashboardPage(self.browser)



from __future__ import absolute_import

from selenium.webdriver.common.by import By

from . import onboarding_students_page
from .teach_base_page import TeachBasePage


class OnboardingClassesPage(TeachBasePage):
    def __init__(self, browser):
        super(OnboardingClassesPage, self).__init__(browser)

        assert self.on_correct_page("onboarding_classes_page")

    def create_class(self, name, classmate_progress):
        self.browser.find_element(By.ID, "id_class_name").send_keys(name)
        if classmate_progress:
            self.browser.find_element(By.ID, "id_classmate_progress").click()

        self._click_create_class_button()

        return onboarding_students_page.OnboardingStudentsPage(self.browser)

    def create_class_empty(self):
        self._click_create_class_button()

        return self

    def _click_create_class_button(self):
        self.browser.find_element(By.ID, "create_class_button").click()

    def does_not_have_classes(self):
        return self.element_does_not_exist_by_id("add_students")



import csv

from common.models import (
    Class,
    DailyActivity,
    DynamicElement,
    School,
    SchoolTeacherInvitation,
    Student,
    Teacher,
    TotalActivity,
    UserProfile,
)
from django.contrib import admin
from django.contrib.auth.admin import UserAdmin as _UserAdmin
from django.contrib.auth.models import User
from django.http import HttpResponse
from import_export.admin import ExportActionMixin

from portal.forms.admin import (
    AdminChangeUserPasswordForm,
    AdminUserCreationForm,
)
from portal.views.api import anonymise


class ClassAdmin(admin.ModelAdmin, ExportActionMixin):
    search_fields = [
        "name",
        "teacher__new_user__first_name",
        "teacher__new_user__last_name",
        "teacher__school__name",
    ]
    list_display = [
        "__str__",
        "teacher",
        "teacher_school",
        "number_of_students",
        "id",
    ]
    autocomplete_fields = ["teacher"]

    def teacher_school(self, obj):
        return obj.teacher.school

    def number_of_students(self, obj):
        return len(obj.students.all())


class SchoolAdmin(admin.ModelAdmin, ExportActionMixin):
    search_fields = ["name", "country", "county"]
    list_filter = ["county", "country"]
    list_display = [
        "__str__",
        "country",
        "county",
        "number_of_teachers",
        "number_of_classes",
    ]

    def number_of_teachers(self, obj):
        return len(obj.teacher_school.all())

    def number_of_classes(self, obj):
        return len(obj.classes()) if obj.classes() else 0


class StudentAdmin(admin.ModelAdmin, ExportActionMixin):
    search_fields = [
        "new_user__first_name",
        "new_user__last_name",
        "new_user__username",
        "class_field__name",
        "class_field__teacher__new_user__first_name",
        "class_field__teacher__new_user__last_name",
        "class_field__teacher__school__name",
    ]
    list_display = [
        "__str__",
        "class_field",
        "class_field_teacher",
        "class_field_school",
        "new_user",
    ]
    readonly_fields = ["user", "new_user", "login_id"]
    autocomplete_fields = ["class_field", "pending_class_request"]

    def class_field_teacher(self, obj):
        if obj.class_field:
            return obj.class_field.teacher
        else:
            return "Independent"

    def class_field_school(self, obj):
        if obj.class_field:
            return obj.class_field.teacher.school
        else:
            return "Independent"


class TeacherAdmin(admin.ModelAdmin, ExportActionMixin):
    search_fields = [
        "new_user__first_name",
        "new_user__last_name",
        "school__name",
        "new_user__username",
    ]
    list_display = [
        "__str__",
        "school",
        "new_user",
        "number_of_classes",
        "is_admin",
    ]
    readonly_fields = ["user", "new_user"]
    autocomplete_fields = ["school", "invited_by"]

    def number_of_classes(self, obj):
        return len(obj.class_teacher.all())


class UserProfileAdmin(admin.ModelAdmin, ExportActionMixin):
    search_fields = [
        "id",
        "user__first_name",
        "user__last_name",
        "user__username",
        "user__date_joined",
    ]
    list_filter = ["user__date_joined"]
    list_display = ["user", "__str__", "is_verified", "type", "id"]
    readonly_fields = ["user"]

    def type(self, obj):
        if hasattr(obj, "student"):
            if obj.student.class_field is None:
                return "INDEPENDENT"
            return "STUDENT"
        return "TEACHER"


class SchoolTeacherInvitationAdmin(admin.ModelAdmin, ExportActionMixin):
    search_fields = [
        "from_teacher__new_user__first_name",
        "from_teacher__new_user__last_name",
        "from_teacher__new_user__email",
        "school__name",
        "invited_teacher_first_name",
        "invited_teacher_last_name",
        "invited_teacher_email",
        "expiry",
        "creation_time",
    ]
    readonly_fields = ["token", "school", "from_teacher"]


class DynamicElementAdmin(admin.ModelAdmin, ExportActionMixin):
    def has_delete_permission(self, request, obj=None):
        return False


class DailyActivityAdmin(admin.ModelAdmin, ExportActionMixin):
    search_fields = ["date"]
    list_display = ["__str__", "date"]


class TotalActivityAdmin(admin.ModelAdmin, ExportActionMixin):
    def has_add_permission(self, request):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


_UserAdmin.list_display += ("date_joined", "id")
_UserAdmin.list_filter += ("date_joined",)


class UserAdmin(_UserAdmin):
    actions = ["anonymise_user", "export_as_csv"]
    add_form = AdminUserCreationForm
    change_password_form = AdminChangeUserPasswordForm

    def anonymise_user(self, request, queryset):
        for user in queryset:
            anonymise(user)

    anonymise_user.short_description = "Anonymise selected users"

    def export_as_csv(self, request, queryset):
        meta = self.model._meta
        field_names = [
            field.name for field in meta.fields if field.name != "password"
        ]

        response = HttpResponse(content_type="text/csv")
        response["Content-Disposition"] = "attachment; filename={}.csv".format(
            meta
        )
        writer = csv.writer(response)

        writer.writerow(field_names)
        for obj in queryset:
            writer.writerow([getattr(obj, field) for field in field_names])

        return response

    export_as_csv.short_description = "Export selected users data as CSV"


admin.site.register(Class, ClassAdmin)
admin.site.register(Student, StudentAdmin)
admin.site.register(Teacher, TeacherAdmin)
admin.site.register(School, SchoolAdmin)
admin.site.unregister(User)
admin.site.register(User, UserAdmin)
admin.site.register(UserProfile, UserProfileAdmin)
admin.site.register(SchoolTeacherInvitation, SchoolTeacherInvitationAdmin)
admin.site.register(DynamicElement, DynamicElementAdmin)
admin.site.register(DailyActivity, DailyActivityAdmin)
admin.site.register(TotalActivity, TotalActivityAdmin)



from django.utils import timezone

from portal.app_settings import SCREENTIME_WARNING_EXPIRY_TIME


class ScreentimeWarningMiddleware:
    """Middleware used to set the milliseconds until the user will see the screentime warning"""

    def __init__(self, get_response):
        self.get_response = get_response
        # One-time configuration and initialization.

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.
        if request.user.is_authenticated:
            if "last_screentime_warning" not in request.session:
                request.session["last_screentime_warning"] = request.user.last_login.timestamp()

            screentime_warning_time = request.session["last_screentime_warning"] + SCREENTIME_WARNING_EXPIRY_TIME

            # Subtract the current time from the screentime warning time and convert to milliseconds
            request.session["screentime_warning_timeout"] = (
                screentime_warning_time - timezone.now().timestamp()
            ) * 1000

        response = self.get_response(request)

        return response



from selenium.webdriver.common.by import By

from . import student_login_page
from .base_page import BasePage


class StudentLoginClassCodePage(BasePage):
    def __init__(self, browser):
        super(StudentLoginClassCodePage, self).__init__(browser)

        assert self.on_correct_page("student_login_page_class_code")

    def student_input_access_code(self, access_code):
        self._student_input_access_code(access_code)

        return student_login_page.StudentLoginPage(self.browser)

    def student_input_access_code_failure(self, access_code):
        self._student_input_access_code(access_code)
        return self

    def _student_input_access_code(self, access_code):
        self.browser.find_element(By.ID, "id_access_code").send_keys(access_code)
        self.browser.find_element(By.NAME, "school_login_class_code").click()

    def has_access_code_input_failed(self, form_id, error):
        errors = self.browser.find_element(By.ID, form_id).find_element(By.CLASS_NAME, "errorlist").text
        return error in errors



from django.shortcuts import render

from portal.strings.play import (
    PLAY_BANNER,
    PLAY_HEADLINE,
)


def play_landing_page(request):
    return render(
        request,
        "portal/play.html",
        {
            "BANNER": PLAY_BANNER,
            "HEADLINE": PLAY_HEADLINE,
        },
    )



from selenium.webdriver.common.by import By


def is_message_showing(browser, message):
    return message in browser.find_element(By.ID, "messages").text


def is_email_verified_message_showing(browser):
    return is_message_showing(browser, "Your email address was successfully verified, please log in.")


def is_teacher_details_updated_message_showing(browser):
    return is_message_showing(browser, "Your account details have been successfully changed.")


def is_student_details_updated_message_showing(browser):
    return is_message_showing(browser, "Your account details have been changed successfully.")


def is_indep_student_join_request_received_message_showing(browser):
    return is_message_showing(browser, "Your request to join a school has been received successfully.")


def is_indep_student_join_request_revoked_message_showing(browser):
    return is_message_showing(browser, "Your request to join a school has been cancelled successfully.")


def is_email_updated_message_showing(browser):
    return is_message_showing(
        browser,
        "Your email will be changed once you have verified it, until then " "you can still log in with your old email.",
    )


def is_password_updated_message_showing(browser):
    return is_message_showing(browser, "Please login using your new password.")


def is_organisation_created_message_showing(browser, name):
    return is_message_showing(browser, "The school or club '%s' has been successfully added." % name)


def is_class_created_message_showing(browser, name):
    return is_message_showing(browser, "The class '%s' has been created successfully." % name)


def is_contact_message_sent_message_showing(browser):
    return is_message_showing(browser, "Your message was sent successfully.")



from __future__ import absolute_import

from selenium.webdriver.common.by import By

from . import class_page
from .teach_base_page import TeachBasePage


class TeachDismissStudentsPage(TeachBasePage):
    def __init__(self, browser):
        super(TeachDismissStudentsPage, self).__init__(browser)

        assert self.on_correct_page("dismiss_students_page")

    def enter_email(self, email, id=0):
        self.browser.find_element(By.ID, f"id_form-{id}-email").send_keys(email)
        self.browser.find_element(By.ID, f"id_form-{id}-confirm_email").send_keys(email)
        return self

    def cancel(self):
        self.browser.find_element(By.ID, "cancel_button").click()
        return class_page.TeachClassPage(self.browser)

    def dismiss(self):
        self.browser.find_element(By.ID, "dismiss_button").click()
        return class_page.TeachClassPage(self.browser)



from datetime import timedelta
from uuid import uuid4

from common.helpers.emails import (
    DotmailerUserType,
    add_to_dotmailer,
    generate_token,
    update_email,
)
from common.helpers.generators import get_random_username
from common.mail import address_book_ids, campaign_ids, send_dotdigital_email
from common.models import (
    Class,
    JoinReleaseStudent,
    SchoolTeacherInvitation,
    Student,
    Teacher,
)
from common.permissions import check_teacher_authorised, logged_in_as_teacher
from common.utils import using_two_factor
from django.contrib import messages as messages
from django.contrib.auth import logout
from django.contrib.auth.decorators import login_required, user_passes_test
from django.contrib.auth.models import User
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse, reverse_lazy
from django.utils import timezone
from django.views.decorators.http import require_POST
from game.level_management import levels_shared_with, unshare_level
from game.models import Level
from two_factor.utils import devices_for_user

from portal.forms.invite_teacher import InviteTeacherForm
from portal.forms.organisation import OrganisationForm
from portal.forms.registration import DeleteAccountForm
from portal.forms.teach import (
    ClassCreationForm,
    InvitedTeacherForm,
    TeacherAddExternalStudentForm,
    TeacherEditAccountForm,
)
from portal.helpers.decorators import ratelimit
from portal.helpers.password import check_update_password
from portal.helpers.ratelimit import (
    RATELIMIT_LOGIN_GROUP,
    RATELIMIT_LOGIN_RATE,
    RATELIMIT_METHOD,
    clear_ratelimit_cache_for_user,
)
from .teach import create_class


def _get_update_account_rate(group, request):
    """
    Custom rate which checks in a POST request is performed on the update
    account form on the teacher dashboard. It needs to check if
    "update_account" is in the POST request because there are 2 other forms
    on the teacher dashboard that can also perform POST request, but we
    do not want to ratelimit those.
    :return: the rate used in the decorator below.
    """
    return RATELIMIT_LOGIN_RATE if "update_account" in request.POST else None


def _get_update_account_ratelimit_key(group, request):
    """
    Get the username from the request as a ratelimit cache key.
    :return: the username from the request.
    """
    return request.user.username


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
@ratelimit(
    group=RATELIMIT_LOGIN_GROUP,
    key=_get_update_account_ratelimit_key,
    method=RATELIMIT_METHOD,
    rate=_get_update_account_rate,
    block=True,
)
def dashboard_teacher_view(request, is_admin):
    teacher = request.user.new_teacher
    school = teacher.school

    coworkers = None
    sent_invites = []
    update_school_form = None

    if school:
        coworkers = Teacher.objects.filter(school=school).order_by(
            "new_user__last_name", "new_user__first_name"
        )

        sent_invites = (
            SchoolTeacherInvitation.objects.filter(school=school)
            if teacher.is_admin
            else []
        )

        update_school_form = OrganisationForm(
            user=request.user, current_school=school
        )
        update_school_form.fields["name"].initial = school.name
        update_school_form.fields["country"].initial = school.country
        update_school_form.fields["county"].initial = school.county

    invite_teacher_form = InviteTeacherForm()

    create_class_form = ClassCreationForm(teacher=teacher)

    update_account_form = TeacherEditAccountForm(request.user)
    update_account_form.fields["first_name"].initial = request.user.first_name
    update_account_form.fields["last_name"].initial = request.user.last_name

    delete_account_form = DeleteAccountForm(request.user)
    delete_account_confirm = False

    anchor = ""

    backup_tokens = check_backup_tokens(request)

    show_onboarding_complete = False

    if request.method == "POST":
        if can_process_update_school_form(request, is_admin):
            anchor = "school-details"
            update_school_form = OrganisationForm(
                request.POST, user=request.user, current_school=school
            )
            anchor = process_update_school_form(request, school, anchor)

        elif "create_class" in request.POST:
            anchor = "new-class"
            create_class_form = ClassCreationForm(request.POST, teacher=teacher)
            if create_class_form.is_valid():
                class_teacher = teacher
                # If the logged in teacher is an admin, then get the class teacher from the selected dropdown
                if teacher.is_admin:
                    class_teacher = get_object_or_404(
                        Teacher, id=create_class_form.cleaned_data["teacher"]
                    )
                created_class = create_class(
                    create_class_form, class_teacher, class_creator=teacher
                )
                messages.success(
                    request,
                    "The class '{className}' has been created successfully.".format(
                        className=created_class.name
                    ),
                )
                return HttpResponseRedirect(
                    reverse_lazy(
                        "view_class",
                        kwargs={"access_code": created_class.access_code},
                    )
                )

        elif request.POST.get("show_onboarding_complete") == "1":
            show_onboarding_complete = True

        elif "invite_teacher" in request.POST:
            invite_teacher_form = InviteTeacherForm(request.POST)
            if invite_teacher_form.is_valid():
                data = invite_teacher_form.cleaned_data
                invited_teacher_first_name = data["teacher_first_name"]
                invited_teacher_last_name = data["teacher_last_name"]
                invited_teacher_email = data["teacher_email"]
                invited_teacher_is_admin = data["make_admin_ticked"]

                token = uuid4().hex
                SchoolTeacherInvitation.objects.create(
                    token=token,
                    school=school,
                    from_teacher=teacher,
                    invited_teacher_first_name=invited_teacher_first_name,
                    invited_teacher_last_name=invited_teacher_last_name,
                    invited_teacher_email=invited_teacher_email,
                    invited_teacher_is_admin=invited_teacher_is_admin,
                    expiry=timezone.now() + timedelta(days=30),
                )

                account_exists = User.objects.filter(
                    email=invited_teacher_email
                ).exists()

                registration_link = f"{request.build_absolute_uri(reverse('invited_teacher', kwargs={'token': token}))} "

                campaign_id = (
                    campaign_ids["invite_teacher_with_account"]
                    if account_exists
                    else campaign_ids["invite_teacher_without_account"]
                )

                send_dotdigital_email(
                    campaign_id,
                    [invited_teacher_email],
                    personalization_values={
                        "SCHOOL_NAME": school.name,
                        "REGISTRATION_LINK": registration_link,
                    },
                )

                messages.success(
                    request,
                    f"You have invited {invited_teacher_first_name} {invited_teacher_last_name} to your school.",
                )

                # Clear form
                invite_teacher_form = InviteTeacherForm()

        elif "delete_account" in request.POST:
            delete_account_form = DeleteAccountForm(request.user, request.POST)
            if not delete_account_form.is_valid():
                messages.warning(
                    request,
                    "Your account was not deleted due to incorrect password.",
                )
            else:
                delete_account_confirm = True
        else:
            anchor = "account"
            update_account_form = TeacherEditAccountForm(
                request.user, request.POST
            )
            (
                changing_email,
                new_email,
                changing_password,
                anchor,
            ) = process_update_account_form(request, teacher, anchor)
            if changing_email:
                logout(request)
                messages.success(
                    request,
                    "Your email will be changed once you have verified it, until then "
                    "you can still log in with your old email.",
                )
                return render(
                    request,
                    "portal/email_verification_needed.html",
                    {"usertype": "TEACHER"},
                )

            if changing_password:
                logout(request)
                messages.success(
                    request, "Please login using your new password."
                )
                return HttpResponseRedirect(reverse_lazy("teacher_login"))

    if teacher.is_admin:
        # Making sure the current teacher classes come up first
        classes = school.classes()
        [
            classes.insert(0, classes.pop(i))
            for i in range(len(classes))
            if classes[i].teacher.id == teacher.id
        ]

        requests = list(
            Student.objects.filter(
                pending_class_request__teacher__school=school
            )
        )
        [
            requests.insert(0, requests.pop(i))
            for i in range(len(requests))
            if requests[i].pending_class_request.teacher.id == teacher.id
        ]

    else:
        classes = Class.objects.filter(teacher=teacher)
        requests = Student.objects.filter(
            pending_class_request__teacher=teacher
        )

    return render(
        request,
        "portal/teach/dashboard.html",
        {
            "teacher": teacher,
            "classes": classes,
            "is_admin": is_admin,
            "coworkers": coworkers,
            "requests": requests,
            "invite_teacher_form": invite_teacher_form,
            "update_school_form": update_school_form,
            "create_class_form": create_class_form,
            "update_account_form": update_account_form,
            "delete_account_form": delete_account_form,
            "delete_account_confirm": delete_account_confirm,
            "anchor": anchor,
            "backup_tokens": backup_tokens,
            "show_onboarding_complete": show_onboarding_complete,
            "sent_invites": sent_invites,
        },
    )


def can_process_update_school_form(request, is_admin):
    return "update_school" in request.POST and is_admin


def check_backup_tokens(request):
    backup_tokens = 0
    # For teachers using 2FA, find out how many backup tokens they have
    if using_two_factor(request.user):
        try:
            backup_tokens = request.user.staticdevice_set.all()[
                0
            ].token_set.count()
        except Exception:
            backup_tokens = 0

    return backup_tokens


def process_update_school_form(request, school, old_anchor):
    update_school_form = OrganisationForm(
        request.POST, user=request.user, current_school=school
    )
    if update_school_form.is_valid():
        data = update_school_form.cleaned_data
        name = data.get("name", "")
        country = data.get("country")
        county = school.county

        school.name = name
        school.country = country
        school.save()

        anchor = "#"

        messages.success(
            request,
            "You have updated the details for your school or club successfully.",
        )
    else:
        anchor = old_anchor

    return anchor


def process_update_account_form(request, teacher, old_anchor):
    update_account_form = TeacherEditAccountForm(request.user, request.POST)
    changing_email = False
    changing_password = False
    new_email = ""
    if update_account_form.is_valid():
        data = update_account_form.cleaned_data

        # check not default value for CharField
        changing_password = check_update_password(
            update_account_form, teacher.new_user, request, data
        )

        teacher.new_user.first_name = data["first_name"]
        teacher.new_user.last_name = data["last_name"]

        changing_email, new_email = update_email(teacher, request, data)

        teacher.save()
        teacher.new_user.save()

        anchor = ""

        # Reset ratelimit cache after successful account details update
        clear_ratelimit_cache_for_user(teacher.new_user.username)

        messages.success(
            request, "Your account details have been successfully changed."
        )
    else:
        anchor = old_anchor

    return changing_email, new_email, changing_password, anchor


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def dashboard_manage(request):
    teacher = request.user.new_teacher

    if teacher.school or request.GET.get("account") == "true":
        return dashboard_teacher_view(request, teacher.is_admin)
    else:
        return HttpResponseRedirect(reverse_lazy("onboarding-organisation"))


def check_teacher_is_authorised(teacher, user):
    if teacher == user or (teacher.school != user.school or not user.is_admin):
        raise Http404


@require_POST
@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def organisation_kick(request, pk):
    teacher = get_object_or_404(Teacher, id=pk)
    user = request.user.new_teacher

    check_teacher_is_authorised(teacher, user)

    success_message = (
        "The teacher has been successfully removed from your school or club."
    )

    classes = Class.objects.filter(teacher=teacher)
    for klass in classes:
        teacher_id = request.POST.get(klass.access_code, None)
        if teacher_id:
            new_teacher = get_object_or_404(Teacher, id=teacher_id)
            klass.teacher = new_teacher
            klass.save()

            success_message = success_message.replace(
                ".", " and their classes were successfully transferred."
            )

    classes = Class.objects.filter(teacher=teacher)
    teachers = Teacher.objects.filter(school=teacher.school).exclude(
        id=teacher.id
    )

    if classes.exists():
        messages.info(
            request,
            "This teacher still has classes assigned to them. You must first move them "
            "to another teacher in your school or club.",
        )
        return render(
            request,
            "portal/teach/teacher_move_all_classes.html",
            {
                "original_teacher": teacher,
                "classes": classes,
                "teachers": teachers,
                "submit_button_text": "Move classes and remove teacher",
            },
        )

    teacher.school = None
    teacher.save()

    messages.success(request, success_message)

    send_dotdigital_email(
        campaign_ids["teacher_released"],
        [teacher.new_user.email],
        personalization_values={"SCHOOL_CLUB_NAME": user.school.name},
    )

    return HttpResponseRedirect(reverse_lazy("dashboard"))


@require_POST
@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def invite_toggle_admin(request, invite_id):
    invite = SchoolTeacherInvitation.objects.filter(id=invite_id)[0]
    invite.invited_teacher_is_admin = not invite.invited_teacher_is_admin
    invite.save()

    if invite.invited_teacher_is_admin:
        messages.success(
            request, "Administrator invite status has been given successfully"
        )
        send_dotdigital_email(
            campaign_ids["admin_given"],
            [invite.invited_teacher_email],
            personalization_values={
                "SCHOOL_CLUB_NAME": invite.school,
                "MANAGEMENT_LINK": request.build_absolute_uri(
                    reverse("dashboard")
                ),
            },
        )

    else:
        messages.success(
            request, "Administrator invite status has been revoked successfully"
        )
        send_dotdigital_email(
            campaign_ids["admin_revoked"],
            [invite.invited_teacher_email],
            personalization_values={"SCHOOL_CLUB_NAME": invite.school},
        )

    return HttpResponseRedirect(reverse_lazy("dashboard"))


@require_POST
@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def organisation_toggle_admin(request, pk):
    teacher = get_object_or_404(Teacher, id=pk)
    user = request.user.new_teacher

    check_teacher_is_authorised(teacher, user)

    teacher.is_admin = not teacher.is_admin
    teacher.save()

    if teacher.is_admin:
        messages.success(
            request, "Administrator status has been given successfully."
        )
        send_dotdigital_email(
            campaign_ids["admin_given"],
            [teacher.new_user.email],
            personalization_values={
                "SCHOOL_CLUB_NAME": teacher.school.name,
                "MANAGEMENT_LINK": request.build_absolute_uri(
                    reverse("dashboard")
                ),
            },
        )
    else:
        # Remove access to all levels that are from other teachers' students
        [
            unshare_level(level, teacher.new_user)
            for level in levels_shared_with(teacher.new_user)
            if hasattr(level.owner, "student")
            and not teacher.teaches(level.owner)
        ]
        messages.success(
            request, "Administrator status has been revoked successfully."
        )
        send_dotdigital_email(
            campaign_ids["admin_revoked"],
            [teacher.new_user.email],
            personalization_values={"SCHOOL_CLUB_NAME": teacher.school.name},
        )

    return HttpResponseRedirect(reverse_lazy("dashboard"))


@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_disable_2FA(request, pk):
    teacher = get_object_or_404(Teacher, id=pk)
    user = request.user.new_teacher

    # check user has authority to change
    if teacher.school != user.school or not user.is_admin:
        raise Http404

    [
        device.delete()
        for device in devices_for_user(teacher.new_user)
        if request.method == "POST"
    ]

    return HttpResponseRedirect(reverse_lazy("dashboard"))


@require_POST
@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_accept_student_request(request, pk):
    student = get_object_or_404(Student, id=pk)

    check_student_request_can_be_handled(request, student)

    students = Student.objects.filter(
        class_field=student.pending_class_request
    ).order_by("new_user__first_name")

    if request.method == "POST":
        form = TeacherAddExternalStudentForm(
            student.pending_class_request, request.POST
        )
        if form.is_valid():
            data = form.cleaned_data
            student.class_field = student.pending_class_request
            teacher = student.pending_class_request.teacher
            student.pending_class_request = None
            student.new_user.username = get_random_username()
            student.new_user.first_name = data["name"]
            student.new_user.last_name = ""
            student.new_user.email = ""

            students_levels = Level.objects.filter(owner=student.new_user.userprofile).all()
            school_admins = teacher.school.admins()
            for level in students_levels:
                level.shared_with.add(*[school_admin.new_user.id for school_admin in school_admins])

                if not teacher.is_admin:
                    level.shared_with.add(teacher.new_user)

                level.needs_approval = True
                level.save()

            student.save()
            student.new_user.save()
            student.new_user.userprofile.save()

            # log the data
            joinrelease = JoinReleaseStudent.objects.create(
                student=student, action_type=JoinReleaseStudent.JOIN
            )
            joinrelease.save()

            return render(
                request,
                "portal/teach/teacher_added_external_student.html",
                {"student": student, "class": student.class_field},
            )
    else:
        form = TeacherAddExternalStudentForm(
            student.pending_class_request,
            initial={"name": student.new_user.first_name},
        )

    return render(
        request,
        "portal/teach/teacher_add_external_student.html",
        {
            "students": students,
            "class": student.pending_class_request,
            "student": student,
            "form": form,
        },
    )


def check_student_request_can_be_handled(request, student):
    """
    Check student is awaiting decision on request
    """
    if not student.pending_class_request:
        raise Http404

    # check user (teacher) has authority to accept student
    check_teacher_authorised(request, student.pending_class_request.teacher)


@require_POST
@login_required(login_url=reverse_lazy("teacher_login"))
@user_passes_test(logged_in_as_teacher, login_url=reverse_lazy("teacher_login"))
def teacher_reject_student_request(request, pk):
    student = get_object_or_404(Student, id=pk)

    check_student_request_can_be_handled(request, student)

    send_dotdigital_email(
        campaign_ids["student_join_request_rejected"],
        [student.new_user.email],
        personalization_values={
            "SCHOOL_CLUB_NAME": student.pending_class_request.teacher.school.name,
            "ACCESS_CODE": student.pending_class_request.access_code,
        },
    )

    student.pending_class_request = None
    student.save()

    messages.success(
        request,
        "Request from external/independent student has been rejected successfully.",
    )

    return HttpResponseRedirect(reverse_lazy("dashboard"))


@login_required(login_url=reverse_lazy("teacher_login"))
def delete_teacher_invite(request, token):
    try:
        invite = SchoolTeacherInvitation.objects.get(token=token)
    except SchoolTeacherInvitation.DoesNotExist:
        invite = None
    teacher = request.user.new_teacher

    # auth the user before deletion
    if invite is None or teacher.school != invite.school:
        messages.error(
            request,
            "You do not have permission to perform this action or the invite does not exist",
        )
    else:
        invite_teacher_first_name = invite.invited_teacher_first_name
        invite.anonymise()
        messages.success(
            request,
            f"Invite for {invite_teacher_first_name} successfully deleted",
        )
    return HttpResponseRedirect(reverse_lazy("dashboard"))


@login_required(login_url=reverse_lazy("teacher_login"))
def resend_invite_teacher(request, token):
    try:
        invite = SchoolTeacherInvitation.objects.get(token=token)
    except SchoolTeacherInvitation.DoesNotExist:
        invite = None
    teacher = request.user.new_teacher

    # auth the user before deletion
    if invite is None or teacher.school != invite.school:
        messages.error(
            request,
            "You do not have permission to perform this action or the invite does not exist",
        )
    else:
        invite.expiry = timezone.now() + timedelta(days=30)
        invite.save()
        teacher = Teacher.objects.filter(id=invite.from_teacher.id)[0]

        messages.success(request, "Teacher re-invited!")

        registration_link = f"{request.build_absolute_uri(reverse('invited_teacher', kwargs={'token': token}))} "

        campaign_id = (
            campaign_ids["invite_teacher_with_account"]
            if teacher.exists()
            else campaign_ids["invite_teacher_without_account"]
        )

        send_dotdigital_email(
            campaign_id,
            [invite.invited_teacher_email],
            personalization_values={
                "SCHOOL_NAME": invite.school,
                "REGISTRATION_LINK": registration_link,
            },
        )

    return HttpResponseRedirect(reverse_lazy("dashboard"))


def invited_teacher(request, token):
    error_message = process_teacher_invitation(request, token)

    if request.method == "POST":
        invited_teacher_form = InvitedTeacherForm(request.POST)
        if invited_teacher_form.is_valid():
            messages.success(
                request,
                "Your account has been created successfully, please log in.",
            )
            return HttpResponseRedirect(reverse_lazy("teacher_login"))
    else:
        invited_teacher_form = InvitedTeacherForm()

    return render(
        request,
        "portal/teach/invited.html",
        {
            "invited_teacher_form": invited_teacher_form,
            "error_message": error_message,
        },
    )


def process_teacher_invitation(request, token):
    try:
        invitation = SchoolTeacherInvitation.objects.get(
            token=token, expiry__gt=timezone.now()
        )
    except SchoolTeacherInvitation.DoesNotExist:
        return "Uh oh, the Invitation does not exist or it has expired. 😞"

    if User.objects.filter(email=invitation.invited_teacher_email).exists():
        return (
            "It looks like an account is already registered with this email address. You will need to delete the "
            "other account first or change the email associated with it in order to proceed. You will then be able to "
            "access this page."
        )
    else:
        if request.method == "POST":
            invited_teacher_form = InvitedTeacherForm(request.POST)
            if invited_teacher_form.is_valid():
                data = invited_teacher_form.cleaned_data
                invited_teacher_password = data["teacher_password"]
                newsletter_ticked = data["newsletter_ticked"]

                # Create the teacher
                invited_teacher = Teacher.objects.factory(
                    first_name=invitation.invited_teacher_first_name,
                    last_name=invitation.invited_teacher_last_name,
                    email=invitation.invited_teacher_email,
                    password=invited_teacher_password,
                )
                invited_teacher.is_admin = invitation.invited_teacher_is_admin
                invited_teacher.school = invitation.school
                invited_teacher.invited_by = invitation.from_teacher
                invited_teacher.save()

                # Verify their email
                generate_token(invited_teacher.new_user, preverified=True)

                # Add to Dotmailer if they ticked the box
                if newsletter_ticked:
                    user = invited_teacher.user.user
                    add_to_dotmailer(
                        user.first_name,
                        user.last_name,
                        user.email,
                        address_book_ids["newsletter"],
                        DotmailerUserType.TEACHER,
                    )

                # Anonymise the invitation
                invitation.anonymise()



from django.conf import settings

CONTACT_FORM_EMAILS = getattr(settings, "PORTAL_CONTACT_FORM_EMAIL", ("codeforlife@ocado.com",))

# Private key for Recaptcha
RECAPTCHA_PRIVATE_KEY = getattr(settings, "RECAPTCHA_PRIVATE_KEY", None)

# Public key for Recaptcha
RECAPTCHA_PUBLIC_KEY = getattr(settings, "RECAPTCHA_PUBLIC_KEY", None)

DEBUG = getattr(settings, "DEBUG", False)

# The permission function for checking if the request is coming from a cron job
IS_CLOUD_SCHEDULER_FUNCTION = getattr(settings, "IS_CLOUD_SCHEDULER_FUNCTION", lambda _: False)

# Half an hour
SESSION_EXPIRY_TIME = 60 * 30

# One hour
SCREENTIME_WARNING_EXPIRY_TIME = 60 * 60



from rest_framework.request import Request
from rest_framework.response import Response

from ..permissions import IsCronRequestFromGoogle


class CronMixin:
    http_method_names = ["get"]
    permission_classes = [IsCronRequestFromGoogle]

    def get(self, request: Request) -> Response:
        raise NotImplementedError()



from django.conf import settings
from rest_framework.permissions import BasePermission
from rest_framework.request import Request
from rest_framework.views import View


class IsCronRequestFromGoogle(BasePermission):
    """
    Validate that requests to your cron URLs are coming from App Engine and not from another source.
    https://cloud.google.com/appengine/docs/flexible/scheduling-jobs-with-cron-yaml#securing_urls_for_cron
    """

    def has_permission(self, request: Request, view: View):
        return settings.DEBUG or request.META.get("HTTP_X_APPENGINE_CRON") == "true"



from __future__ import absolute_import

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By

from .base_page import BasePage


class PasswordResetPage(BasePage):
    def __init__(self, browser):
        super(PasswordResetPage, self).__init__(browser)

        assert self.on_correct_page("reset_password_form_page")

    def teacher_reset_password(self, new_password):
        self.clear_and_fill("new_password1", new_password)
        self.clear_and_fill("new_password2", new_password)

        self.browser.find_element(By.ID, "teacher_update_button").click()

        self.wait_for_element_by_id("reset_password_done_page")

        return self

    def student_reset_password(self, new_password):
        self.clear_and_fill("new_password1", new_password)
        self.clear_and_fill("new_password2", new_password)

        self.browser.find_element(By.ID, "student_update_button").click()

        self.wait_for_element_by_id("reset_password_done_page")

        return self

    def reset_password_fail(self, new_password):
        self.clear_and_fill("new_password1", new_password)
        self.clear_and_fill("new_password2", new_password)

        try:
            update_button = self.browser.find_element(By.ID, "teacher_update_button")
        except NoSuchElementException:
            update_button = self.browser.find_element(By.ID, "student_update_button")

        update_button.click()

        return self

    def clear_and_fill(self, field, value):
        self.browser.find_element(By.ID, "id_" + field).clear()
        self.browser.find_element(By.ID, "id_" + field).send_keys(value)



from __future__ import unicode_literals

from django.db import models, migrations
from django.conf import settings

from django.contrib.auth.hashers import make_password
from django.utils import timezone
import os

# Functions from the following migrations need manual copying.
# Move them and any dependencies into this file, then update the
# RunPython operations to refer to the local versions:
# portal.migrations.0001_squashed_0036_data_viewing_user_fix
# portal.migrations.0037_userprofile_developer
# portal.migrations.0041_new_news
# portal.migrations.0040_initial_news


def insert_admin_user(apps, schema_editor):
    User = apps.get_model("auth", "User")
    admin = User.objects.create()
    admin.username = "portaladmin"
    admin.email = ("codeforlife-portal@ocado.com",)
    admin.is_superuser = True
    admin.is_staff = True
    admin.password = make_password(os.getenv("ADMIN_PASSWORD", "abc123"))
    admin.save()


def insert_users(apps, schema_editor):
    User = apps.get_model("auth", "User")
    UserProfile = apps.get_model("portal", "UserProfile")
    School = apps.get_model("portal", "School")
    Teacher = apps.get_model("portal", "Teacher")
    Class = apps.get_model("portal", "Class")
    Student = apps.get_model("portal", "Student")

    # Create school
    school = School.objects.create(
        name="Swiss Federal Polytechnic",
        postcode="AL10 9NE",
        town="Welwyn Hatfield",
        latitude="51.76183",
        longitude="-0.244361",
    )

    # Create teachers' users
    teacher_user = User.objects.create(
        username="test teacher",
        first_name="Albert",
        last_name="Einstein",
        email="alberteinstein@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    teacher2_user = User.objects.create(
        username="test teacher2",
        first_name="Max",
        last_name="Planck",
        email="maxplanck@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    teacher3_user = User.objects.create(
        username="media ram",
        first_name="Ram",
        last_name="Leith",
        email="ramleith@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    # Create students' users
    student1_user = User.objects.create(
        username="test student1",
        first_name="Leonardo",
        last_name="DaVinci",
        email="leonardodavinci@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student2_user = User.objects.create(
        username="test student2",
        first_name="Galileo",
        last_name="Galilei",
        email="galileogalilei@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student3_user = User.objects.create(
        username="Issac",
        first_name="Isaac",
        last_name="Newton",
        email="isaacnewton@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student4_user = User.objects.create(
        username="test student4",
        first_name="Richard",
        last_name="Feynman",
        email="richardfeynman@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student5_user = User.objects.create(
        username="test student5",
        first_name="Alexander",
        last_name="Flemming",
        email="alexanderflemming@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student6_user = User.objects.create(
        username="test student6",
        first_name="Daniel",
        last_name="Bernoulli",
        email="danielbernoulli@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    indy_user = User.objects.create(
        username="indy",
        first_name="Indiana",
        last_name="Jones",
        email="indianajones@codeforlife.com",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student7_user = User.objects.create(
        username="media noah",
        first_name="Noah",
        last_name="Monaghan",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student8_user = User.objects.create(
        username="media elliot",
        first_name="Elliot",
        last_name="Sharp",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student9_user = User.objects.create(
        username="media tajmae",
        first_name="Tajmae",
        last_name="Joseph",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student10_user = User.objects.create(
        username="media carlton",
        first_name="Carlton",
        last_name="Joseph",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student11_user = User.objects.create(
        username="media nadal",
        first_name="Nadal",
        last_name="Spencer-Jennings",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student12_user = User.objects.create(
        username="media freddie",
        first_name="Freddie",
        last_name="Goff",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student13_user = User.objects.create(
        username="media leon",
        first_name="Leon",
        last_name="Scott",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    student14_user = User.objects.create(
        username="media betty",
        first_name="Betty",
        last_name="Kessell",
        password=make_password(os.getenv("ADMIN_PASSWORD", "Password1")),
    )

    # Create UserProfiles
    teacher_userprofile = UserProfile.objects.create(user=teacher_user, developer=True)
    teacher2_userprofile = UserProfile.objects.create(user=teacher2_user)
    teacher3_userprofile = UserProfile.objects.create(user=teacher3_user)
    student1_userprofile = UserProfile.objects.create(
        user=student1_user, developer=True
    )
    student2_userprofile = UserProfile.objects.create(user=student2_user)
    student3_userprofile = UserProfile.objects.create(user=student3_user)
    student4_userprofile = UserProfile.objects.create(user=student4_user)
    student5_userprofile = UserProfile.objects.create(user=student5_user)
    student6_userprofile = UserProfile.objects.create(user=student6_user)
    indy_userprofile = UserProfile.objects.create(user=indy_user)
    student7_userprofile = UserProfile.objects.create(user=student7_user)
    student8_userprofile = UserProfile.objects.create(user=student8_user)
    student9_userprofile = UserProfile.objects.create(user=student9_user)
    student10_userprofile = UserProfile.objects.create(user=student10_user)
    student11_userprofile = UserProfile.objects.create(user=student11_user)
    student12_userprofile = UserProfile.objects.create(user=student12_user)
    student13_userprofile = UserProfile.objects.create(user=student13_user)
    student14_userprofile = UserProfile.objects.create(user=student14_user)

    # Create teachers
    teacher = Teacher.objects.create(
        title="Mr",
        user=teacher_userprofile,
        school=school,
        is_admin=True,
        pending_join_request=None,
    )
    teacher2 = Teacher.objects.create(
        title="Mr",
        user=teacher2_userprofile,
        school=school,
        is_admin=False,
        pending_join_request=None,
    )
    teacher3 = Teacher.objects.create(
        title="Mrs",
        user=teacher3_userprofile,
        school=school,
        is_admin=True,
        pending_join_request=None,
    )

    # Create classes
    klass = Class.objects.create(
        name="Class 101",
        teacher=teacher,
        access_code="AB123",
        classmates_data_viewable=True,
        always_accept_requests=True,
    )
    class2 = Class.objects.create(
        name="Class 102",
        teacher=teacher2,
        access_code="AB124",
        classmates_data_viewable=True,
        always_accept_requests=True,
    )

    class3 = Class.objects.create(
        name="Class 103",
        teacher=teacher2,
        access_code="AB125",
        classmates_data_viewable=True,
        always_accept_requests=True,
    )

    class4 = Class.objects.create(
        name="Young Coders 101",
        teacher=teacher3,
        access_code="RL123",
        classmates_data_viewable=True,
        always_accept_requests=True,
    )

    # Create students
    student1 = Student.objects.create(
        class_field=klass, user=student1_userprofile, pending_class_request=None
    )

    student2 = Student.objects.create(
        class_field=klass, user=student2_userprofile, pending_class_request=None
    )

    student3 = Student.objects.create(
        class_field=None, user=student3_userprofile, pending_class_request=None
    )

    student4 = Student.objects.create(
        class_field=class2, user=student4_userprofile, pending_class_request=None
    )

    student5 = Student.objects.create(
        class_field=class2, user=student5_userprofile, pending_class_request=None
    )

    student6 = Student.objects.create(
        class_field=class3, user=student6_userprofile, pending_class_request=None
    )

    indy_student = Student.objects.create(
        user=indy_userprofile, pending_class_request=None
    )

    student7 = Student.objects.create(
        class_field=class4, user=student7_userprofile, pending_class_request=None
    )

    student8 = Student.objects.create(
        class_field=class4, user=student8_userprofile, pending_class_request=None
    )

    student9 = Student.objects.create(
        class_field=class4, user=student9_userprofile, pending_class_request=None
    )

    student10 = Student.objects.create(
        class_field=class4, user=student10_userprofile, pending_class_request=None
    )

    student11 = Student.objects.create(
        class_field=class4, user=student11_userprofile, pending_class_request=None
    )

    student12 = Student.objects.create(
        class_field=class4, user=student12_userprofile, pending_class_request=None
    )

    student13 = Student.objects.create(
        class_field=class4, user=student13_userprofile, pending_class_request=None
    )

    student14 = Student.objects.create(
        class_field=class4, user=student14_userprofile, pending_class_request=None
    )

    user = User.objects.create(
        username="DATA_AGGREGATE",
        first_name="DATA",
        last_name="AGGREGATE",
        email="aggregator@codeforlife.com",
        password=make_password(os.getenv("DATA_AGGREGATE_PASSWORD", "Password1")),
    )

    user_profile = UserProfile.objects.create(user=user, can_view_aggregated_data=True)


def insert_news(apps, schema_editor):
    FrontPageNews = apps.get_model("portal", "FrontPageNews")

    news1 = FrontPageNews.objects.create(
        title="Teachers 'not confident on coding'",
        text="Computer coding is being introduced to the school curriculum in less than six weeks' time, but many primary school teachers will be under-prepared to teach the new subject, according to a new poll.",
        link="http://www.dailymail.co.uk/wires/pa/article-2700124/TEACHERS-NOT-CONFIDENT-ON-CODING.html?ITO=1490&ns_mchannel=rss&ns_campaign=1490",
        link_text="Read more on www.dailymail.co.uk",
        added_dstamp=timezone.now(),
    )
    news2 = FrontPageNews.objects.create(
        title="British schools are not prepared to teach coding",
        text="With just six weeks until the new computing curriculum is introduced in UK schools, research has revealed that British primary school teachers are not fully prepared to teach their pupils how to code.",
        link="http://www.cbronline.com/news/social/british-teachers-are-not-prepared-to-teach-coding-4322259",
        link_text="Read more on www.cbronline.co.uk",
        added_dstamp=timezone.now(),
    )
    news3 = FrontPageNews.objects.create(
        title="Ocado launches free Code For Life tool to help 130,000 apprehensive primary school teachers",
        text="Thousands of primary school teachers aren’t ready to teach computer coding lessons that become part of the country’s national curriculum for computing before the start of the academic year in September.",
        link="http://www.itproportal.com/2014/07/21/ocado-launches-free-code-life-tool-help-130000-apprehensive-primary-school-teachers/?utm_source=feedburner&utm_medium=feed&utm_campaign=Feed%3A+itproportal%2Frss+(Latest+ITProPortal+News)",
        link_text="Read more on itproportal.com",
        added_dstamp=timezone.now(),
    )
    news4 = FrontPageNews.objects.create(
        title="Ocado Technology launches coding resources for teachers",
        text="Ocado Technology is the latest business to get involved with preparing teachers for the new computing curriculum that comes into effect in September",
        link="http://www.computerworlduk.com/news/careers/3531070/ocado-technology-launches-coding-resources-for-teachers/",
        link_text="Read more on www.computerworlduk.com",
        added_dstamp=timezone.now(),
    )
    news5 = FrontPageNews.objects.create(
        title="Not Enough Teachers Will Know Code By September",
        text="A new curriculum brings coding into primary schools in September – and it looks as if teachers won’t be able to cope. ",
        link="http://www.techweekeurope.co.uk/news/not-enough-teachers-know-code-149462",
        link_text="Read more on www.techweekeurope.co.uk",
        added_dstamp=timezone.now(),
    )
    news6 = FrontPageNews.objects.create(
        title="Ocado Technology readies primary school teachers with code initiative",
        text="Ocado Technology has launched a coding initiative after finding that 73% of primary school teachers feel they have not been given the necessary resources to teach children to code.",
        link="http://www.computerweekly.com/news/2240225117/Ocado-Technology-readies-primary-school-teachers-with-code-initiative",
        link_text="Read more on www.computerweekly.com",
        added_dstamp=timezone.now(),
    )
    news7 = FrontPageNews.objects.create(
        title="Curriculum countdown",
        text="According to new research more than 130,000 primary school teachers don’t feel confident enough to teach computer coding",
        link="http://edtechnology.co.uk/Featured-Content/curriculum_countdown",
        link_text="Read more on www.edtechnology.co.uk",
        added_dstamp=timezone.now(),
    )
    news8 = FrontPageNews.objects.create(
        title="Computing Curriculum Countdown: Over 130,000 Primary School Teachers Don’t Feel Confident Enough to Teach Computer Coding",
        text="Many primary school teachers feel they haven’t been given the necessary resources to teach the new Computing curriculum from September. ",
        link="http://www.primarytimes.net/news/2014/07/computing-curriculum-countdown-over-130-000-primary-school-teachers-don-t-feel-confident-enough-to-teach-computer-coding-",
        link_text="Read more on www.primarytimes.net",
        added_dstamp=timezone.now(),
    )
    news9 = FrontPageNews.objects.create(
        title="Ocado’s technology team to release primary school coding tool",
        text="Ocado’s technology team is launching Code for Life, the online grocer’s initiative to help children learn to code.",
        link="http://internetretailing.net/2014/07/ocados-technology-team-launch-primary-school-coding-tool/",
        link_text="Read more on www.internetretailing.net",
        added_dstamp=timezone.now(),
    )
    news10 = FrontPageNews.objects.create(
        title="Teachers 'not confident' over coding",
        text="Computer coding is being introduced to the school curriculum in less than six weeks' time, but many primary school teachers will be under-prepared to teach the new subject, according to a new poll.",
        link="http://www.dailyecho.co.uk/leisure/technology/11358091.Teachers__not_confident__over_coding/",
        link_text="Read more on www.dailyecho.co.uk",
        added_dstamp=timezone.now(),
    )
    news11 = FrontPageNews.objects.create(
        title="Ocado Technology launches coding resources for teachers",
        text="Ocado Technology is a latest business to get concerned with scheming teachers for a new computing curriculum that comes into outcome in September.",
        link="http://www.datacentremanagement.org/2014/07/ocado-technology-launches-coding-resources-for-teachers/",
        link_text="Read more on www.datacentremanagement.org",
        added_dstamp=timezone.now(),
    )
    news12 = FrontPageNews.objects.create(
        title="Teachers 'not confident enough' to teach code",
        text="More than 70% of primary school teachers say they don't feel confident enough to teach the new coding syllabus being introduced this September.",
        link="http://www.newelectronics.co.uk/electronics-news/teachers-not-confident-enough-to-teach-code/62697/",
        link_text="Read more on www.newelectronics.co.uk",
        added_dstamp=timezone.now(),
    )


def insert_news2(apps, schema_editor):
    FrontPageNews = apps.get_model("portal", "FrontPageNews")

    news1 = FrontPageNews.objects.create(
        title="Bletchley Park: From Code-Breaking to Kids Coding",
        text="The National Museum of Computer has opened its free Weekend Codability Project. This is part of the Code for Life initiative which aims to inspire the next generation of computer scientists.",
        link="http://www.idgconnect.com/blog-abstract/9056/bletchley-park-from-code-breaking-kids-coding",
        link_text="Read more on idgconnect.com",
        added_dstamp=timezone.now(),
    )
    news2 = FrontPageNews.objects.create(
        title="Giving kids Codability",
        text="Weekend Codability aims to empower young people by introducing them to programming computers. Children will be taught how to give instructions to computers, change existing instructions in programs and create their own programs.",
        link="http://edtechnology.co.uk/News/giving_kids_codability",
        link_text="Read more on edtechnology.co.uk",
        added_dstamp=timezone.now(),
    )
    news3 = FrontPageNews.objects.create(
        title="Try your hand at Bletchley Park codability project",
        text="Following the introduction of computing to England’s school curriculum last month, young people across the country are being invited to try their hand at programming computers in Block H, the world's first purpose-built computer centre, on Bletchley Park.",
        link="http://www.mkweb.co.uk/COMPUTERS-Try-hand-Bletchley-Park-codability/story-23860128-detail/story.html",
        link_text="Read more on mkweb.co.uk",
        added_dstamp=timezone.now(),
    )


class Migration(migrations.Migration):

    replaces = [
        ("portal", "0001_initial"),
        ("portal", "0002_admin_user"),
        ("portal", "0003_school_teacher_relationship"),
        ("portal", "0004_class_access_code"),
        ("portal", "0005_student_pin"),
        ("portal", "0006_teacher_email_verification"),
        ("portal", "0007_school_admin"),
        ("portal", "0008_auto_20140729_0938"),
        ("portal", "0009_remove_student_pin"),
        ("portal", "0010_auto_20140729_1030"),
        ("portal", "0009_teacher_pending_join_request"),
        ("portal", "0011_merge"),
        ("portal", "0012_auto_20140729_1359"),
        ("portal", "0013_auto_20140730_1100"),
        ("portal", "0014_auto_20140730_1459"),
        ("portal", "0015_student_pending_class_request"),
        ("portal", "0016_emailverification_email"),
        ("portal", "0017_school_postcode"),
        ("portal", "0018_auto_20140804_1246"),
        ("portal", "0019_auto_20140804_1713"),
        ("portal", "0020_auto_20140806_1632"),
        ("portal", "0020_auto_20140806_1205"),
        ("portal", "0021_merge"),
        ("portal", "0022_remove_student_name"),
        ("portal", "0023_class_classmates_data_viewable"),
        ("portal", "0024_auto_20140813_1536"),
        ("portal", "0025_trial_users"),
        ("portal", "0026_fix_trial_users"),
        ("portal", "0027_more_trial_users"),
        ("portal", "0028_demoting_max_planck"),
        ("portal", "0029_trial_independent_student"),
        ("portal", "0030_media_users"),
        ("portal", "0031_auto_20140903_1450"),
        ("portal", "0032_userprofile_can_view_aggregated_data"),
        ("portal", "0033_data_viewing_user"),
        ("portal", "0034_data_viewing_user_fix"),
        ("portal", "0035_data_viewing_user_hack"),
        ("portal", "0036_data_viewing_user_fix"),
        ("portal", "0037_userprofile_developer"),
        ("portal", "0038_frontpagenews"),
        ("portal", "0039_auto_20141109_1827"),
        ("portal", "0040_initial_news"),
        ("portal", "0041_new_news"),
    ]

    dependencies = [
        ("auth", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="UserProfile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "avatar",
                    models.ImageField(
                        default="static/portal/img/avatars/default-avatar.jpeg",
                        null=True,
                        upload_to="static/portal/img/avatars/",
                        blank=True,
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE
                    ),
                ),
                ("awaiting_email_verification", models.BooleanField(default=False)),
                ("can_view_aggregated_data", models.BooleanField(default=False)),
                ("developer", models.BooleanField(default=False)),
            ],
        ),
        migrations.CreateModel(
            name="School",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("postcode", models.CharField(max_length=10)),
                ("latitude", models.CharField(max_length=20)),
                ("longitude", models.CharField(max_length=20)),
                ("town", models.CharField(max_length=200)),
            ],
        ),
        migrations.CreateModel(
            name="Teacher",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        to="portal.UserProfile", on_delete=models.CASCADE
                    ),
                ),
                (
                    "pending_join_request",
                    models.ForeignKey(
                        related_name="join_request",
                        to="portal.School",
                        null=True,
                        on_delete=models.SET_NULL,
                    ),
                ),
                ("is_admin", models.BooleanField(default=False)),
                ("title", models.CharField(max_length=35)),
                (
                    "school",
                    models.ForeignKey(
                        related_name="teacher_school",
                        to="portal.School",
                        null=True,
                        on_delete=models.SET_NULL,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Class",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("access_code", models.CharField(max_length=5)),
                ("classmates_data_viewable", models.BooleanField(default=False)),
                ("accept_requests_until", models.DateTimeField(null=True)),
                ("always_accept_requests", models.BooleanField(default=False)),
                (
                    "teacher",
                    models.ForeignKey(
                        related_name="class_teacher",
                        to="portal.Teacher",
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
            options={"verbose_name_plural": "classes"},
        ),
        migrations.CreateModel(
            name="Student",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "class_field",
                    models.ForeignKey(
                        related_name="students",
                        to="portal.Class",
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "user",
                    models.OneToOneField(
                        to="portal.UserProfile", on_delete=models.CASCADE
                    ),
                ),
                (
                    "pending_class_request",
                    models.ForeignKey(
                        related_name="class_request",
                        to="portal.Class",
                        null=True,
                        on_delete=models.CASCADE,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Guardian",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("children", models.ManyToManyField(to="portal.Student")),
                (
                    "user",
                    models.OneToOneField(
                        to="portal.UserProfile", on_delete=models.CASCADE
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EmailVerification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("token", models.CharField(max_length=30)),
                ("expiry", models.DateTimeField()),
                ("used", models.BooleanField(default=False)),
                (
                    "user",
                    models.ForeignKey(
                        related_name="email_verifications",
                        to="portal.UserProfile",
                        on_delete=models.CASCADE,
                    ),
                ),
                (
                    "email",
                    models.CharField(
                        default=None, max_length=200, null=True, blank=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FrontPageNews",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                ("title", models.CharField(max_length=200)),
                ("text", models.CharField(max_length=1000)),
                ("link", models.CharField(max_length=500)),
                ("link_text", models.CharField(max_length=200)),
                ("added_dstamp", models.DateTimeField()),
            ],
        ),
        migrations.RunPython(code=insert_admin_user),
        migrations.RunPython(code=insert_users),
        migrations.RunPython(code=insert_news),
        migrations.RunPython(code=insert_news2),
    ]



from common.models import Teacher
from django.shortcuts import render
from two_factor.views import LoginView
from two_factor.forms import AuthenticationTokenForm, BackupTokenForm

from portal.forms.teach import TeacherLoginForm
from portal.views.home import redirect_teacher_to_correct_page
from . import has_user_lockout_expired

# This import is required so that 2FA works properly
from portal import handlers


class TeacherLoginView(LoginView):
    template_name = "portal/login/teacher.html"
    form_list = (
        ("auth", TeacherLoginForm),
        ("token", AuthenticationTokenForm),
        ("backup", BackupTokenForm),
    )

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or redirect_teacher_to_correct_page(self.request, self.request.user.userprofile.teacher)

    def post(self, request, *args, **kwargs):
        """
        If the email address inputted in the form corresponds to that of a blocked
        account, this redirects the user to the locked out page. However, if the lockout
        time is more than 24 hours before this is executed, the account is unlocked.
        """
        wizard_step = self.request.POST.get("teacher_login_view-current_step", None)

        if wizard_step == "auth":
            email = request.POST.get("auth-username")
            if Teacher.objects.filter(new_user__email=email).exists():
                teacher = Teacher.objects.get(new_user__email=email)

                if teacher.blocked_time is not None:
                    if has_user_lockout_expired(teacher):
                        teacher.blocked_time = None
                        teacher.save()
                    else:
                        return render(
                            self.request,
                            "portal/locked_out.html",
                            {"is_teacher": True},
                        )

        return super(TeacherLoginView, self).post(request, *args, **kwargs)



from datetime import timedelta

import jwt
from common.models import Teacher
from common.permissions import logged_in_as_independent_student
from django.conf import settings
from django.contrib import messages as messages
from django.contrib.auth.models import User
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.urls import reverse_lazy
from django.utils import timezone


def verify_email(request, token):
    decoded_jwt = jwt.decode(token, settings.SECRET_KEY, algorithms=["HS256"])

    user_found = User.objects.filter(email=decoded_jwt["email"]).first()
    usertype = (
        "TEACHER"
        if Teacher.objects.filter(new_user=user_found).exists()
        else "INDEP_STUDENT"
    )
    is_expired = decoded_jwt["expires"] < timezone.now().timestamp()

    is_changing_email = decoded_jwt["new_email"] != ""
    updated_email = (
        decoded_jwt["new_email"] if is_changing_email else decoded_jwt["email"]
    )

    if not user_found or is_expired:
        return render(
            request,
            "portal/email_verification_failed.html",
            {"usertype": usertype},
        )

    is_user_verified = user_found.userprofile.is_verified
    if is_user_verified and not is_changing_email:
        return render(
            request,
            "portal/email_verification_failed.html",
            {"usertype": usertype},
        )

    if not is_user_verified or is_changing_email:
        user_found.email = updated_email
        user_found.username = updated_email
        user_found.userprofile.is_verified = True
        user_found.userprofile.save()
        user_found.save()

    messages.success(
        request, "Your email address was successfully verified, please log in."
    )

    if logged_in_as_independent_student(user_found):
        login_url = "independent_student_login"
    else:
        login_url = "teacher_login"

    return HttpResponseRedirect(reverse_lazy(login_url))


def has_verification_failed(verifications):
    return (
        len(verifications) > 1
        or verifications[0].verified
        or (verifications[0].expiry - timezone.now()) < timedelta()
    )
